# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Bbox utils"""
import json
import math
import itertools as it
import numpy as np
from easydict import EasyDict as ed

ssd300_config = ed({
    "img_shape": [300, 300],
    "num_ssd_boxes": 1917,
    "match_threshold": 0.5,
    "nms_threshold": 0.6,
    "min_score": 0.7,
    "max_boxes": 100,

    # learing rate settings
    "lr_init": 0.001,
    "lr_end_rate": 0.001,
    "warmup_epochs": 2,
    "momentum": 0.9,
    "weight_decay": 1.5e-4,

    # network
    "num_default": [3, 6, 6, 6, 6, 6],
    "extras_in_channels": [256, 576, 1280, 512, 256, 256],
    "extras_out_channels": [576, 1280, 512, 256, 256, 128],
    "extras_strides": [1, 1, 2, 2, 2, 2],
    "extras_ratio": [0.2, 0.2, 0.2, 0.25, 0.5, 0.25],
    "feature_size": [19, 10, 5, 3, 2, 1],
    "min_scale": 0.2,
    "max_scale": 0.95,
    "aspect_ratios": [(), (2, 3), (2, 3), (2, 3), (2, 3), (2, 3)],
    "steps": (16, 32, 64, 100, 150, 300),
    "prior_scaling": (0.1, 0.2),
    "gamma": 2.0,
    "alpha": 0.75,
})


class GenerateDefaultBoxes():
    """
    Generate Default boxes for SSD300, follows the order of (W, H, archor_sizes).
    `self.default_boxes` has a shape of [archor_sizes, H, W, 4], the last dimension is [y, x, h, w].
    `self.default_boxes_tlbr` has a shape as `self.default_boxes`, the last dimension is [y1, x1, y2, x2].
    """

    def __init__(self):
        fk = ssd300_config.img_shape[0] / np.array(ssd300_config.steps)
        scale_rate = (ssd300_config.max_scale - ssd300_config.min_scale) / (len(ssd300_config.num_default) - 1)
        scales = [ssd300_config.min_scale + scale_rate * i for i in range(len(ssd300_config.num_default))] + [1.0]
        self.default_boxes = []
        for idex, feature_size in enumerate(ssd300_config.feature_size):
            sk1 = scales[idex]
            sk2 = scales[idex + 1]
            sk3 = math.sqrt(sk1 * sk2)
            if idex == 0 and not ssd300_config.aspect_ratios[idex]:
                w, h = sk1 * math.sqrt(2), sk1 / math.sqrt(2)
                all_sizes = [(0.1, 0.1), (w, h), (h, w)]
            else:
                all_sizes = [(sk1, sk1)]
                for aspect_ratio in ssd300_config.aspect_ratios[idex]:
                    w, h = sk1 * math.sqrt(aspect_ratio), sk1 / math.sqrt(aspect_ratio)
                    all_sizes.append((w, h))
                    all_sizes.append((h, w))
                all_sizes.append((sk3, sk3))

            assert len(all_sizes) == ssd300_config.num_default[idex]

            for i, j in it.product(range(feature_size), repeat=2):
                for w, h in all_sizes:
                    cx, cy = (j + 0.5) / fk[idex], (i + 0.5) / fk[idex]
                    self.default_boxes.append([cy, cx, h, w])

        def to_tlbr(cy, cx, h, w):
            return cy - h / 2, cx - w / 2, cy + h / 2, cx + w / 2

        # For IoU calculation
        self.default_boxes_tlbr = np.array(tuple(to_tlbr(*i) for i in self.default_boxes), dtype='float32')
        self.default_boxes = np.array(self.default_boxes, dtype='float32')


ssd_default_boxes_tlbr = GenerateDefaultBoxes().default_boxes_tlbr
ssd_default_boxes = GenerateDefaultBoxes().default_boxes


def ssd_bboxes_encode(boxes):
    r"""
    Labels anchors with ground truth inputs.

    Args:
        boxes (numpy.ndarray): Ground truth with shape [N, 5], for each row, it stores
            [ymin, xmin, ymax, xmax, cls].

    Returns:
        numpy.ndarray, location ground truth with shape [num_anchors, 4].
        numpy.ndarray, class ground truth with shape [num_anchors, 1].
        numpy.ndarray, number of positives in an image.
    """
    y1, x1, y2, x2 = np.split(ssd_default_boxes_tlbr[:, :4], 4, axis=-1)
    vol_anchors = (x2 - x1) * (y2 - y1)

    def jaccard_with_anchors(bbox):
        """Compute jaccard score a box and the anchors."""
        # Intersection bbox and volume.
        ymin = np.maximum(y1, bbox[0])
        xmin = np.maximum(x1, bbox[1])
        ymax = np.minimum(y2, bbox[2])
        xmax = np.minimum(x2, bbox[3])
        w = np.maximum(xmax - xmin, 0.)
        h = np.maximum(ymax - ymin, 0.)

        # Volumes.
        inter_vol = h * w
        union_vol = vol_anchors + (bbox[2] - bbox[0]) * (bbox[3] - bbox[1]) - inter_vol
        jaccard = inter_vol / union_vol
        return np.squeeze(jaccard)

    pre_scores = np.zeros((ssd300_config.num_ssd_boxes), dtype=np.float32)
    t_boxes = np.zeros((ssd300_config.num_ssd_boxes, 4), dtype=np.float32)
    t_label = np.zeros((ssd300_config.num_ssd_boxes), dtype=np.int64)
    for bbox in boxes:
        # Add one for inserting background label
        label = int(bbox[4]) + 1
        scores = jaccard_with_anchors(bbox)
        idx = np.argmax(scores)
        scores[idx] = 2.0
        mask = (scores > ssd300_config.match_threshold)
        mask = mask & (scores > pre_scores)
        pre_scores = np.maximum(pre_scores, scores * mask)
        t_label = mask * label + (1 - mask) * t_label
        for i in range(4):
            t_boxes[:, i] = mask * bbox[i] + (1 - mask) * t_boxes[:, i]

    index = np.nonzero(t_label)

    # Transform to tlbr.
    bboxes = np.zeros((ssd300_config.num_ssd_boxes, 4), dtype=np.float32)
    bboxes[:, [0, 1]] = (t_boxes[:, [0, 1]] + t_boxes[:, [2, 3]]) / 2
    bboxes[:, [2, 3]] = t_boxes[:, [2, 3]] - t_boxes[:, [0, 1]]

    # Encode features.
    bboxes_t = bboxes[index]
    default_boxes_t = ssd_default_boxes[index]
    bboxes_t[:, :2] = (bboxes_t[:, :2] - default_boxes_t[:, :2]) / \
        (default_boxes_t[:, 2:] * ssd300_config.prior_scaling[0])
    tmp = np.maximum(bboxes_t[:, 2:4] / default_boxes_t[:, 2:4], 0.000001)
    bboxes_t[:, 2:4] = np.log(tmp) / ssd300_config.prior_scaling[1]
    bboxes[index] = bboxes_t

    num_match = np.array([len(np.nonzero(t_label)[0])], dtype=np.int32)
    return bboxes, t_label.astype(np.int32), num_match


def ssd_bboxes_filter(boxes, box_scores, image_shape):
    r"""
    Filter predict boxes with minimum score and nms threshold.

    Args:
        boxes (numpy.ndarray): Ground truth with shape [N, 4], for each row, it stores
            [ymin, xmin, ymax, xmax].
        box_scores (numpy.ndarray): Class scores with shape [N, 21].
        image_shape (tuple): Shape of original image with the format [h, w].

    Returns:
        list[list[float]], ground truth with shape [N, 4], for each row, it stores
        [ymin, xmin, ymax, xmax].
        list[list[float]], class scores with shape [N, 21].
        list[list[int]], class label with shape [N, 21].
    """
    final_boxes = []
    final_label = []
    final_score = []
    h, w = image_shape
    # Ignore background(0) label class
    for c in range(1, box_scores.shape[1]):
        class_box_scores = box_scores[:, c]
        score_mask = class_box_scores > ssd300_config.min_score
        class_box_scores = class_box_scores[score_mask]
        class_boxes = boxes[score_mask] * [h, w, h, w]

        if score_mask.any():
            nms_index = apply_nms(class_boxes, class_box_scores,
                                  ssd300_config.nms_threshold)
            class_boxes = class_boxes[nms_index]
            class_box_scores = class_box_scores[nms_index]

            final_boxes += class_boxes.tolist()
            final_score += class_box_scores.tolist()
            final_label += [c] * len(class_box_scores)

    return final_boxes, final_score, final_label


def jaccard_numpy(box_a, box_b):
    """Compute the jaccard overlap of two sets of boxes."""
    def intersect(box_a, box_b):
        """Compute the intersect of two sets of boxes."""
        max_yx = np.minimum(box_a[:, 2:4], box_b[2:4])
        min_yx = np.maximum(box_a[:, :2], box_b[:2])
        inter = np.clip((max_yx - min_yx), a_min=0, a_max=np.inf)
        return inter[:, 0] * inter[:, 1]

    inter = intersect(box_a, box_b)
    area_a = ((box_a[:, 2] - box_a[:, 0]) *
              (box_a[:, 3] - box_a[:, 1]))
    area_b = ((box_b[2] - box_b[0]) *
              (box_b[3] - box_b[1]))
    union = area_a + area_b - inter
    return inter / union


def apply_nms(all_boxes, all_scores, thres=0.6, max_boxes=100):
    """Apply NMS to all bounding boxes."""
    y1 = all_boxes[:, 0]
    x1 = all_boxes[:, 1]
    y2 = all_boxes[:, 2]
    x2 = all_boxes[:, 3]
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)

    order = all_scores.argsort()[::-1]
    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        if len(keep) >= max_boxes:
            break

        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])
        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h

        ovr = inter / (areas[i] + areas[order[1:]] - inter)
        inds = np.where(ovr <= thres)[0]
        order = order[inds + 1]
    return keep


def coco_eval(pred_data, anno_file):
    """Calculate mAP of predicted bboxes."""
    from pycocotools.coco import COCO
    from pycocotools.cocoeval import COCOeval

    predictions = []
    img_ids = []
    for sample in pred_data:
        pred_boxes = sample['boxes']
        box_scores = sample['box_scores']
        img_id = sample['img_id']
        img_ids.append(img_id)

        final_pred = ssd_bboxes_filter(pred_boxes, box_scores, sample['image_shape'])

        for loc, score, label in zip(final_pred[0], final_pred[1], final_pred[2]):
            res = {}
            res['image_id'] = img_id
            res['bbox'] = [loc[1], loc[0], loc[3] - loc[1], loc[2] - loc[0]]
            res['score'] = score
            res['category_id'] = label
            predictions.append(res)
    with open('predictions.json', 'w') as f:
        json.dump(predictions, f)

    coco_gt = COCO(anno_file)
    coco_dt = coco_gt.loadRes('predictions.json')
    E = COCOeval(coco_gt, coco_dt, iouType='bbox')
    E.params.imgIds = img_ids
    E.evaluate()
    E.accumulate()
    E.summarize()
    return E.stats[0]



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
from mindspore.dataset.vision.py_transforms import Grayscale, RandomHorizontalFlip as PILRandomHorizontalFlip
from mindspore.dataset.vision.c_transforms import *
from mindspore.dataset.transforms.c_transforms import *

vision_trans = [
    'AutoContrast',
    'BoundingBoxAugment',
    'CenterCrop',
    'CutMixBatch',
    'CutOut',
    'Decode',
    'Equalize',
    'Grayscale',
    'HWC2CHW',
    'Invert',
    'MixUpBatch',
    'Normalize',
    'Pad',
    'PILRandomHorizontalFlip',
    'RandomAffine',
    'RandomColor',
    'RandomColorAdjust',
    'RandomCrop',
    'RandomCropDecodeResize',
    'RandomCropWithBBox',
    'RandomHorizontalFlip',
    'RandomHorizontalFlipWithBBox',
    'RandomPosterize',
    'RandomResize',
    'RandomResizedCrop',
    'RandomResizedCropWithBBox',
    'RandomResizeWithBBox',
    'RandomRotation',
    'RandomSelectSubpolicy',
    'RandomSharpness',
    'RandomSolarize',
    'RandomVerticalFlip',
    'RandomVerticalFlipWithBBox',
    'Rescale',
    'Resize',
    'ResizeWithBBox',
    'SoftDvppDecodeRandomCropResizeJpeg',
    'SoftDvppDecodeResizeJpeg',
    'UniformAugment',
]

common_trans = [
    'Compose',
    'Concatenate',
    'Duplicate',
    'Fill',
    'Mask',
    'OneHot',
    'PadEnd',
    'RandomApply',
    'RandomChoice',
    'Slice',
    'TypeCast',
    'Unique',
]

__all__ = vision_trans + common_trans

decode = Decode()
hwc2chw = HWC2CHW()

__all__.extend([
    'decode',
    'hwc2chw',
])



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
from tinyms import layers
from tinyms import ParameterTuple

from tinyms.context import ParallelMode, get_auto_parallel_context
from tinyms.layers import DistributedGradReducer
from tinyms.optimizers import Adam
from tinyms.primitives import HyperMap, GradOperation
from tinyms.primitives import depend
from tinyms.primitives import DType, Fill, ReduceMean, ReduceSum, Shape, Square
from tinyms.primitives import SigmoidCrossEntropyWithLogits
from mindspore.communication.management import get_group_size


class DeepFMWithLoss(layers.Layer):
    """
    Provide DeepFM training loss through network.

    Args:
        network (layers.Layer): The training network.
        l2_coef (float): value for l2 loss. Default: 1e-6.

    Returns:
        Tensor, the loss of the network.

    Examples:
        >>> from tinyms.model import deepfm, DeepFMWithLoss, DeepFMTrainModel
        >>>
        >>> net = deepfm()
        >>> train_net = DeepFMTrainModel(DeepFMWithLoss(net))
    """
    def __init__(self, network, l2_coef=1e-6):
        super(DeepFMWithLoss, self).__init__(auto_prefix=False)
        self.network = network
        self.l2_coef = l2_coef
        self.square = Square()
        self.reduce_mean = ReduceMean(keep_dims=False)
        self.reduce_sum = ReduceSum(keep_dims=False)
        self.loss = SigmoidCrossEntropyWithLogits()

    def construct(self, ids, wts, labels):
        predicts, fm_id_weight, fm_id_embeds = self.network(ids, wts)
        log_loss = self.loss(predicts, labels)
        mean_log_loss = self.reduce_mean(log_loss)
        l2_loss_w = self.reduce_sum(self.square(fm_id_weight))
        l2_loss_v = self.reduce_sum(self.square(fm_id_embeds))
        l2_loss_all = self.l2_coef * (l2_loss_v + l2_loss_w) * 0.5
        loss_df = mean_log_loss + l2_loss_all
        return loss_df


class DeepFMTrainModel(layers.Layer):
    """
    Provide DeepFM training network.

    Args:
        network (layers.Layer): The base network.
        learning_rate (float): A value or a graph for the learning rate. Default: 0.0005.
        eps (float): Term added to the denominator to improve numerical stability.
        Should be greater than 0. Default: 0.00000005.
        loss_scale (float): A floating point value for the loss scale.
        Should be greater than 0. Default: 1024.0.

    Returns:
        Tensor, the value passed by last operator.

    Examples:
        >>> from tinyms.model import deepfm, DeepFMWithLoss, DeepFMTrainModel
        >>>
        >>> net = deepfm()
        >>> train_net = DeepFMTrainModel(DeepFMWithLoss(net))
    """
    def __init__(self, network, learning_rate=0.0005, eps=0.00000005, loss_scale=1024.0):
        super(DeepFMTrainModel, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_train()
        self.weights = ParameterTuple(network.trainable_params())
        self.optimizer = Adam(self.weights, learning_rate=learning_rate, eps=eps, loss_scale=loss_scale)
        self.hyper_map = HyperMap()
        self.grad = GradOperation(get_by_list=True, sens_param=True)
        self.sens = loss_scale

        self.reducer_flag = False
        self.grad_reducer = None
        parallel_mode = get_auto_parallel_context("parallel_mode")
        if parallel_mode in (ParallelMode.DATA_PARALLEL, ParallelMode.HYBRID_PARALLEL):
            self.reducer_flag = True
        if self.reducer_flag:
            mean = get_auto_parallel_context("gradients_mean")
            degree = get_group_size()
            self.grad_reducer = DistributedGradReducer(self.optimizer.parameters, mean, degree)

    def construct(self, ids, wts, labels):
        weights = self.weights
        loss = self.network(ids, wts, labels)
        sens = Fill()(DType()(loss), Shape()(loss), self.sens)
        grads = self.grad(self.network, weights)(ids, wts, labels, sens)
        if self.reducer_flag:
            # apply grad reducer on grads
            grads = self.grad_reducer(grads)
        return depend(loss, self.optimizer(grads))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import os
import cv2

__all__ = ['save_image', 'draw_boxes_on_image', 'visualize_boxes_on_image']

IMG_FORMAT = ('jpg', 'jpeg', 'png')


def save_image(img, save_dir='./', img_name='no_name', img_format='jpg'):
    r"""
    Save the prediction image.

    Args:
        img (numpy.ndarray): the input image.
        save_dir (str): the dir to save the prediction image.
        img_name (str): the name of the prediction image. Default: 'no_name'.
        img_format (str): the format of the prediction image. Default: 'jpg'.
    """
    if img_format.lower() not in IMG_FORMAT:
        raise Exception("当前图片格式仅支持", IMG_FORMAT)
    output_image = os.path.join(save_dir, '{}.{}'.format(img_name, img_format))
    cv2.imwrite(output_image, img)


def draw_boxes_on_image(img, boxes, box_scores, box_classes, box_color=(0, 255, 0),
                        box_thickness=3, text_font=cv2.FONT_HERSHEY_PLAIN,
                        font_scale=3, text_color=(0, 0, 255), font_size=3, show_scores=True):
    r"""
    Draw the prediction box for the input image.

    Args:
        img (numpy.ndarray): the input image.
        boxes (list): the box coordinates.
        box_color (list): the box color. Default: (0, 255, 0).
        box_thickness (int): box thickness. Default: 3.
        text_font (Enum): text font. Default: cv2.FONT_HERSHEY_PLAIN.
        font_scale (int): font scale. Default: 3.
        text_color (list): text color. Default: (0, 0, 255).
        font_size (int): font size. Default: 3.
        show_scores (bool): whether to show scores. Default: True.

    Returns:
        numpy.ndarray, the output image drawed the prediction box.
    """
    x = int(boxes[0])
    y = int(boxes[1])
    w = int(boxes[2])
    h = int(boxes[3])
    cv2.rectangle(img, (x, y), (x+w, y+h), box_color, box_thickness)
    text = '{}, {}'.format(box_classes, str(round(box_scores, 3))) if show_scores else box_classes
    cv2.putText(img, text, (x, y), text_font, font_scale, text_color, font_size)
    return img


def visualize_boxes_on_image(img, bbox_data, box_color=(0, 255, 0),
                             box_thickness=3, text_font=cv2.FONT_HERSHEY_PLAIN,
                             font_scale=3, text_color=(0, 0, 255), font_size=3, show_scores=True):
    r"""
    Visualize the prediction image.

    Args:
        img (numpy.ndarray): the input image.
        bbox_data (dict): the predictions box data.
        box_color (list): the box color. Default: (0, 255, 0).
        box_thickness (int): box thickness. Default: 3.
        text_font (Enum): text font. Default: cv2.FONT_HERSHEY_PLAIN.
        font_scale (int): font scale. Default: 3.
        text_color (list): text color. Default: (0, 0, 255).
        font_size (int): font size. Default: 3.
        show_scores (bool): whether to show scores. Default: True.

    Returns:
        numpy.ndarray, the output image drawed the prediction box.
    """
    bbox_num = len(bbox_data)
    if bbox_num:
        for i in range(bbox_num):
            img = draw_boxes_on_image(img,
                                      bbox_data[i]['bbox'],
                                      bbox_data[i]['score'],
                                      bbox_data[i]['category_id'],
                                      box_color, box_thickness,
                                      text_font,
                                      font_scale,
                                      text_color,
                                      font_size,
                                      show_scores)
    return img



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
from tinyms import layers
from tinyms.primitives import Sigmoid


class DeepFMEvalModel(layers.Layer):
    """
    Provide DeepFM training network.

    Args:
        network (layers.Layer): The base network.

    Returns:
        logits (Tensor) - With the shape as [batch_size, vocab_size] and data type as float32.
        predict_probs (Tensor) - With the same shape as [batch_size, vocab_size] and data type as float32.
        labels (Tensor) - With the same shape as labels.

    Examples:
        >>> from tinyms.model import deepfm, DeepFMEvalModel
        >>>
        >>> net = deepfm()
        >>> eval_net = DeepFMEvalModel(net)
    """
    def __init__(self, network):
        super(DeepFMEvalModel, self).__init__(auto_prefix=False)
        self.network = network
        self.sigmoid = Sigmoid()

    def construct(self, ids, wts, labels):
        logits, _, _ = self.network(ids, wts)
        predict_probs = self.sigmoid(logits)
        return logits, predict_probs, labels



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import tempfile
from tinyms.utils.train import load_checkpoint as _load_checkpoint, \
    load_param_into_net, save_checkpoint as _save_checkpoint

from .hubconf import MODEL_HUB
from .utils.download import download_file_from_url
from .utils.check import HubAssetInfo


def _get_hub_root():
    """
    Get the absolute path of hub package.

    Returns:
        str, return a string of path.
    """
    return os.path.abspath(os.path.dirname(__file__))


def _get_model_asset_path(uid_info):
    """Get model asset yaml name from UidInfo."""
    return os.path.join(_get_hub_root(), 'assets', uid_info.provider, uid_info.version,
                        uid_info.model_info + '.yaml')


def _load_weights(path):
    """
    Load network weights from local asset yaml file.

    This function is used to check whether the network can be loaded normally.

    Args:
        path (str): The path of asset file.
    """
    if not isinstance(path, str):
        raise TypeError(f'`path` must be a string, but got {type(path)}')
    if not os.path.exists(path):
        raise ValueError('Please make sure input is a path.')

    try:
        asset_info = HubAssetInfo(path)
        with tempfile.TemporaryDirectory() as target_path:
            download_url = asset_info.asset['asset-link']
            asset_sha256 = asset_info.asset["asset-sha256"]
            ckpt_path = download_file_from_url(download_url, asset_sha256, target_path)
            return _load_checkpoint(ckpt_path)
    except Exception as e:
        raise Exception(e)


def load(uid, pretrained=True, **kwargs):
    '''
    Load a model from remote TinyMS Hub.

    Args:
        uid (str): Uid. The format should be strictly consistent with
            the official example: `tinyms/0.2/lenet5_v1_mnist`.
        pretrained (bool): Specified if to load pretrained weight ckpt file. Default: True.
        kwargs (dict, optional): Keyword arguments for network initialization.

    Return:
        layers.Layer, the initialized network instance.

    Examples:
        >>> from tinyms import hub
        >>>
        >>> hub.load('tinyms/0.2/lenet5_v1_mnist', class_num=10)
    '''
    uid_info = UidInfo(uid)
    model_info = uid_info.model_name + '_' + uid_info.model_version
    net_func = MODEL_HUB.get(model_info)
    if net_func is None:
        raise ValueError("Currently model_name only supports " + str(list(MODEL_HUB.keys())) + "!")

    net = net_func(**kwargs)
    if pretrained is True:
        asset_path = _get_model_asset_path(uid_info)
        ckpt_params = _load_weights(asset_path)
        load_param_into_net(net, ckpt_params)

    return net


def load_checkpoint(uid, dst, pretrained=True, **kwargs):
    '''
    Load model checkpoint file from remote TinyMS Hub.

    Args:
        uid (str): Uid. The format should be strictly consistent with
            the official example: `tinyms/0.2/lenet5_v1_mnist`.
        dst (str): Full path of filename where the checkpoint file
            will be loaded, e.g. `/tmp/lenet5.ckpt`.
        pretrained (bool): Specified if to load pretrained weight ckpt file. Default: True.
        kwargs (dict, optional): Keyword arguments for network initialization.

    Examples:
        >>> from tinyms import hub
        >>>
        >>> hub.load_checkpoint('tinyms/0.2/lenet5_v1_mnist', '/tmp/lenet5.ckpt', class_num=10)
    '''
    uid_info = UidInfo(uid)
    model_info = uid_info.model_name + '_' + uid_info.model_version
    net_func = MODEL_HUB.get(model_info)
    if net_func is None:
        raise ValueError("Currently model_name only supports " + str(list(MODEL_HUB.keys())) + "!")

    net = net_func(**kwargs)
    if pretrained is True:
        asset_path = _get_model_asset_path(uid_info)
        ckpt_params = _load_weights(asset_path)
        load_param_into_net(net, ckpt_params)
    _save_checkpoint(net, dst)


def load_weights(uid):
    '''
    Load model pretrained weights from remote TinyMS Hub.

    Args:
        uid (str): Uid. The format should be strictly consistent with
            the official example: `tinyms/0.2/lenet5_v1_mnist`.

    Return:
        Dict, the pretrained network weight dict.

    Examples:
        >>> from tinyms import hub
        >>> from tinyms.model import lenet5
        >>> from tinyms.utils.train import load_param_into_net
        >>>
        >>> param_dict = hub.load_weights('tinyms/0.2/lenet5_v1_mnist')
        >>> net = lenet5()
        >>> load_param_into_net(net, param_dict)
    '''
    uid_info = UidInfo(uid)
    asset_path = _get_model_asset_path(uid_info)
    return _load_weights(asset_path)


class UidInfo:
    def __init__(self, uid):
        uid_slice = uid.split('/')
        if len(uid_slice) != 3:
            raise ValueError('The format of uid is not correct! \
                An example for the format is: tinyms/0.2/lenet5_v1_mnist')
        self.provider = uid_slice[0]
        self.version = uid_slice[1]
        self.model_info = uid_slice[2]

        model_slice = self.model_info.split('_')
        if len(uid_slice) != 3:
            raise ValueError('The model format of uid is not correct! \
                An example for the format is: tinyms/0.2/lenet5_v1_mnist')
        self.model_name = model_slice[0]
        self.model_version = model_slice[1]
        self.train_dataset = model_slice[2]

    def __str__(self):
        model_info = '_'.join([self.model_name, self.model_version, self.train_dataset])
        uid_info = '/'.join([self.provider, self.version, model_info])
        return uid_info



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""ResNet Generator."""
from tinyms import layers
from tinyms.primitives import Tanh
from .common_net import ConvNormReLU, ConvTransposeNormReLU


class ResidualBlock(layers.Layer):
    """
    ResNet residual block definition.

    Args:
        dim (int): Input and output channel.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".
        dropout (bool): Use dropout or not. Default: False.
        pad_mode (str): Specifies padding mode. The optional values are "CONSTANT", "REFLECT", "SYMMETRIC".
            Default: "CONSTANT".

    Returns:
        Tensor, output tensor.
    """

    def __init__(self, dim, norm_mode='batch', dropout=True, pad_mode="CONSTANT"):
        super(ResidualBlock, self).__init__()
        self.conv1 = ConvNormReLU(dim, dim, 3, 1, 0, norm_mode, pad_mode)
        self.conv2 = ConvNormReLU(dim, dim, 3, 1, 0, norm_mode, pad_mode, use_relu=False)
        self.dropout = dropout
        if dropout:
            self.dropout = layers.Dropout(0.5)

    def construct(self, x):
        out = self.conv1(x)
        if self.dropout:
            out = self.dropout(out)
        out = self.conv2(out)
        return x + out


class ResNetGenerator(layers.Layer):
    """
    ResNet Generator of GAN.

    Args:
        in_planes (int): Input channel.
        ngf (int): Output channel.
        n_layers (int): The number of ConvNormReLU blocks.
        alpha (float): LeakyRelu slope. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".
        dropout (bool): Use dropout or not. Default: False.
        pad_mode (str): Specifies padding mode. The optional values are "CONSTANT", "REFLECT", "SYMMETRIC".
            Default: "CONSTANT".

    Returns:
        Tensor, output tensor.
    """
    def __init__(self, in_planes=3, ngf=64, n_layers=9, alpha=0.2, norm_mode='batch', dropout=True,
                 pad_mode="CONSTANT"):
        super(ResNetGenerator, self).__init__()
        self.conv_in = ConvNormReLU(in_planes, ngf, 7, 1, alpha=alpha, norm_mode=norm_mode, pad_mode=pad_mode)
        self.down_1 = ConvNormReLU(ngf, ngf * 2, 3, 2, alpha, norm_mode)
        self.down_2 = ConvNormReLU(ngf * 2, ngf * 4, 3, 2, alpha, norm_mode)
        layer_list = [ResidualBlock(ngf * 4, norm_mode, dropout=dropout, pad_mode=pad_mode)] * n_layers
        self.residuals = layers.SequentialLayer(layer_list)
        self.up_2 = ConvTransposeNormReLU(ngf * 4, ngf * 2, 3, 2, alpha, norm_mode)
        self.up_1 = ConvTransposeNormReLU(ngf * 2, ngf, 3, 2, alpha, norm_mode)
        if pad_mode == "CONSTANT":
            self.conv_out = layers.Conv2d(ngf, 3, kernel_size=7, stride=1, pad_mode='pad', padding=3)
        else:
            pad = layers.Pad(paddings=((0, 0), (0, 0), (3, 3), (3, 3)), mode=pad_mode)
            conv = layers.Conv2d(ngf, 3, kernel_size=7, stride=1, pad_mode='pad')
            self.conv_out = layers.SequentialLayer([pad, conv])
        self.activate = Tanh()

    def construct(self, x):
        x = self.conv_in(x)
        x = self.down_1(x)
        x = self.down_2(x)
        x = self.residuals(x)
        x = self.up_2(x)
        x = self.up_1(x)
        output = self.conv_out(x)
        return self.activate(output)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Bert Init."""

from .bert_for_pretraining import BertNetworkWithLoss, BertPreTraining, \
    BertPretrainingLoss, GetMaskedLMOutput, GetNextSentenceOutput, \
    BertTrainOneStepCell, BertTrainOneStepWithLossScaleCell, \
    BertTrainAccumulationAllReduceEachWithLossScaleLayer, \
    BertTrainAccumulationAllReducePostWithLossScaleLayer, \
    BertTrainOneStepWithLossScaleCellForAdam
from .bert import BertAttention, BertEncoderLayer, Bert, bert, \
    BertOutput, BertSelfAttention, BertTransformer, EmbeddingLookup, \
    EmbeddingPostprocessor, RelaPosEmbeddingsGenerator, RelaPosMatrixGenerator, \
    SaturateCast, CreateAttentionMaskFromInputMask
from .bert_for_finetune import BertFinetuneLayer, BertCLS

__all__ = [
    "BertNetworkWithLoss", "BertPreTraining", "BertPretrainingLoss",
    "GetMaskedLMOutput", "GetNextSentenceOutput", "BertTrainOneStepCell",
    "BertTrainOneStepWithLossScaleCell", "BertTrainAccumulationAllReduceEachWithLossScaleLayer",
    "BertTrainAccumulationAllReducePostWithLossScaleLayer",
    "BertAttention", "BertEncoderLayer", "Bert", "bert", "BertOutput",
    "BertSelfAttention", "BertTransformer", "EmbeddingLookup",
    "EmbeddingPostprocessor", "RelaPosEmbeddingsGenerator",
    "RelaPosMatrixGenerator", "SaturateCast", "CreateAttentionMaskFromInputMask",
    "BertTrainOneStepWithLossScaleCellForAdam", "BertFinetuneLayer", "BertCLS"
]



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Losses module. Loss function in machine learning is the target of the model.
It shows how well the model works on a dataset and the optimization target
which the optimizer is searching.
"""
from mindspore.nn import loss
from mindspore.nn.loss import *
from mindspore.nn.loss.loss import _Loss
import tinyms as ts
from . import layers, primitives as P, Tensor
from .model import SSD300


__all__ = [
    'net_with_loss',
    'SSD300WithLoss',
    'CrossEntropyWithLabelSmooth',
    'CycleGANGeneratorLoss',
    'CycleGANDiscriminatorLoss',
]
__all__.extend(loss.__all__)


class SigmoidFocalClassificationLoss(layers.Layer):
    """"
    Sigmoid focal-loss for classification.

    Args:
        gamma (float): Hyper-parameter to balance the easy and hard examples. Default: 2.0
        alpha (float): Hyper-parameter to balance the positive and negative example. Default: 0.25

    Returns:
        Tensor, the focal loss.
    """

    def __init__(self, gamma=2.0, alpha=0.25):
        super(SigmoidFocalClassificationLoss, self).__init__()
        self.sigmiod_cross_entropy = P.SigmoidCrossEntropyWithLogits()
        self.sigmoid = P.Sigmoid()
        self.pow = P.Pow()
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, ts.float32)
        self.off_value = Tensor(0.0, ts.float32)
        self.gamma = gamma
        self.alpha = alpha

    def construct(self, logits, label):
        label = self.onehot(label, P.shape(logits)[-1], self.on_value, self.off_value)
        sigmiod_cross_entropy = self.sigmiod_cross_entropy(logits, label)
        sigmoid = self.sigmoid(logits)
        label = P.cast(label, ts.float32)
        p_t = label * sigmoid + (1 - label) * (1 - sigmoid)
        modulating_factor = self.pow(1 - p_t, self.gamma)
        alpha_weight_factor = label * self.alpha + (1 - label) * (1 - self.alpha)
        focal_loss = modulating_factor * alpha_weight_factor * sigmiod_cross_entropy
        return focal_loss


class SSD300WithLoss(layers.Layer):
    r"""
    Provide SSD300 training loss through network.

    Args:
        network (layers.Layer): The training network.

    Returns:
        Tensor, the loss of the network.

    Examples:
        >>> from tinyms.model import ssd300
        >>> from tinyms.losses import SSD300WithLoss
        >>>
        >>> net = SSD300WithLoss(ssd300())
    """

    def __init__(self, network):
        super(SSD300WithLoss, self).__init__()
        self.network = network
        self.less = P.Less()
        self.tile = P.Tile()
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.expand_dims = P.ExpandDims()
        self.class_loss = SigmoidFocalClassificationLoss(2.0, 0.75)
        self.loc_loss = SmoothL1Loss()

    def construct(self, x, gt_loc, gt_label, num_matched_boxes):
        pred_loc, pred_label = self.network(x)
        mask = P.cast(self.less(0, gt_label), ts.float32)
        num_matched_boxes = self.reduce_sum(P.cast(num_matched_boxes, ts.float32))

        # Localization Loss
        mask_loc = self.tile(self.expand_dims(mask, -1), (1, 1, 4))
        smooth_l1 = self.loc_loss(pred_loc, gt_loc) * mask_loc
        loss_loc = self.reduce_sum(self.reduce_mean(smooth_l1, -1), -1)

        # Classification Loss
        loss_cls = self.class_loss(pred_label, gt_label)
        loss_cls = self.reduce_sum(loss_cls, (1, 2))

        return self.reduce_sum((loss_cls + loss_loc) / num_matched_boxes)


def net_with_loss(net):
    r'''
    This function is provided for AI beginners who are not familiar with which loss
    should be chosen for the network to be trained. Instead of choosing different loss
    function, users could directly get the best suitable loss function by specifying
    network.

    Args:
        net (layers.Layer): The instance of network to be trained.

    Raises:
        TypeError: When network type is not supported.

    Note:
        Currently this function only supports few networks, if the network type is
        not supported, the system would raise TypeError exception.

    Examples:
        >>> from tinyms.model import ssd300
        >>> from tinyms.losses import net_with_loss
        >>>
        >>> net = ssd300()
        >>> net_loss = net_with_loss(net)
    '''
    if not isinstance(net, layers.Layer):
        raise TypeError("Input should be inheritted from layers.Layer!")

    if isinstance(net, SSD300):
        return SSD300WithLoss(net)
    else:
        raise TypeError("Input should be in [SSD300], got {}.".format(type(net)))


class CrossEntropyWithLabelSmooth(_Loss):
    """
    CrossEntropyWith LabelSmooth.

    Args:
        smooth_factor (float): Smooth factor. Default is 0.
        num_classes (int): Number of classes. Default is 1000.

    Returns:
        None.

    Examples:
        >>> CrossEntropyWithLabelSmooth(smooth_factor=0., num_classes=1000)
    """

    def __init__(self, smooth_factor=0., num_classes=1000):
        super(CrossEntropyWithLabelSmooth, self).__init__()
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0 - smooth_factor, ts.float32)
        self.off_value = Tensor(1.0 * smooth_factor /
                                (num_classes - 1), ts.float32)
        self.ce = SoftmaxCrossEntropyWithLogits()
        self.mean = P.ReduceMean(False)
        self.cast = P.Cast()

    def construct(self, logit, label):
        one_hot_label = self.onehot(self.cast(label, ts.int32), P.shape(logit)[1],
                                    self.on_value, self.off_value)
        out_loss = self.ce(logit, one_hot_label)
        out_loss = self.mean(out_loss, 0)
        return out_loss


class GANLoss(_Loss):
    """
    Cycle GAN loss factory.

    Args:
        mode (str): The type of GAN objective. It currently supports 'vanilla', 'lsgan'. Default: 'lsgan'.
        reduction (str): Specifies the reduction to be applied to the output.
            Its value must be one of 'none', 'mean', 'sum'. Default: 'mean'.

    Outputs:
        Tensor or Scalar, if `reduction` is 'none', then output is a tensor and has the same shape as `inputs`.
        Otherwise, the output is a scalar.

    Raises:
        NotImplementedError: Raised when GANLoss mode not recognized.
    """

    def __init__(self, mode="lsgan", reduction='mean'):
        super(GANLoss, self).__init__()
        self.loss = None
        self.ones = P.OnesLike()
        if mode == "lsgan":
            self.loss = loss.MSELoss(reduction)
        elif mode == "vanilla":
            self.loss = BCEWithLogits(reduction)
        else:
            raise NotImplementedError(f'GANLoss {mode} not recognized, we support lsgan and vanilla.')

    def construct(self, predict, target):
        target = P.cast(target, P.dtype(predict))
        target = self.ones(predict) * target
        loss = self.loss(predict, target)
        return loss


class CycleGANGeneratorLoss(_Loss):
    """
    Cycle GAN generator loss.

    Args:
        generator (layers.Layer): Generator of CycleGAN.
        D_A (layers.Layer): The discriminator network of domain A to domain B.
        D_B (layers.Layer): The discriminator network of domain B to domain A.

    Outputs:
        Tuple Tensor, the losses of generator.
    """

    def __init__(self, generator, D_A, D_B):
        super(CycleGANGeneratorLoss, self).__init__()
        self.lambda_A = 10.0
        self.lambda_B = 10.0
        self.lambda_idt = 0.5
        self.use_identity = True
        self.dis_loss = GANLoss("lsgan")
        self.rec_loss = loss.L1Loss("mean")
        self.generator = generator
        self.D_A = D_A
        self.D_B = D_B
        self.true = Tensor(True, ts.bool_)

    def construct(self, img_A, img_B):
        """If use_identity, identity loss will be used."""
        fake_A, fake_B, rec_A, rec_B, identity_A, identity_B = self.generator(img_A, img_B)
        loss_G_A = self.dis_loss(self.D_B(fake_B), self.true)
        loss_G_B = self.dis_loss(self.D_A(fake_A), self.true)
        loss_C_A = self.rec_loss(rec_A, img_A) * self.lambda_A
        loss_C_B = self.rec_loss(rec_B, img_B) * self.lambda_B
        if self.use_identity:
            loss_idt_A = self.rec_loss(identity_A, img_A) * self.lambda_A * self.lambda_idt
            loss_idt_B = self.rec_loss(identity_B, img_B) * self.lambda_B * self.lambda_idt
        else:
            loss_idt_A = 0
            loss_idt_B = 0
        loss_G = loss_G_A + loss_G_B + loss_C_A + loss_C_B + loss_idt_A + loss_idt_B
        return (fake_A, fake_B, loss_G, loss_G_A, loss_G_B, loss_C_A, loss_C_B, loss_idt_A, loss_idt_B)


class CycleGANDiscriminatorLoss(_Loss):
    """
    Cycle GAN discriminator loss.

    Args:
        D_A (layers.Layer): The discriminator network of domain A to domain B.
        D_B (layers.Layer): The discriminator network of domain B to domain A.
        reduction (str): The discriminator network of reduction. Default: none.

    Outputs:
        the loss of discriminator.
    """

    def __init__(self, D_A, D_B, reduction='none'):
        super(CycleGANDiscriminatorLoss, self).__init__()
        self.D_A = D_A
        self.D_B = D_B
        self.false = Tensor(False, ts.bool_)
        self.true = Tensor(True, ts.bool_)
        self.dis_loss = GANLoss("lsgan")
        self.rec_loss = loss.L1Loss("mean")
        self.reduction = reduction

    def construct(self, img_A, img_B, fake_A, fake_B):
        D_fake_A = self.D_A(fake_A)
        D_img_A = self.D_A(img_A)
        D_fake_B = self.D_B(fake_B)
        D_img_B = self.D_B(img_B)
        loss_D_A = self.dis_loss(D_fake_A, self.false) + self.dis_loss(D_img_A, self.true)
        loss_D_B = self.dis_loss(D_fake_B, self.false) + self.dis_loss(D_img_B, self.true)
        loss_D = (loss_D_A + loss_D_B) * 0.5
        return loss_D



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Callback related classes and functions in model training phase."""
import time
import numpy as np
import math
from mindspore.train import callback
from mindspore.train.callback import *
from . import Tensor

__all__ = ['LossTimeMonitor', 'LossTimeMonitorV2', 'BertLossCallBack']
__all__.extend(callback.__all__)


class LossTimeMonitor(Callback):
    """
    Monitor loss and time.

    Args:
        lr_init (numpy.ndarray): Train learning rate. Default: None.

    Returns:
        None

    Examples:
        >>> from tinyms import Tensor
        >>> from tinyms.callbacks import LossTimeMonitor
        >>>
        >>> LossTimeMonitor(lr_init=Tensor([0.05] * 100).asnumpy())
    """

    def __init__(self, lr_init=None):
        super(LossTimeMonitor, self).__init__()
        self.lr_init = lr_init
        self.lr_init_len = len(lr_init)

    def epoch_begin(self, run_context):
        self.losses = []
        self.epoch_time = time.time()

    def epoch_end(self, run_context):
        cb_params = run_context.original_args()

        epoch_mseconds = (time.time() - self.epoch_time) * 1000
        per_step_mseconds = epoch_mseconds / cb_params.batch_num
        print("epoch time: {:5.3f}, per step time: {:5.3f}, avg loss: {:5.3f}".format(epoch_mseconds,
                                                                                      per_step_mseconds,
                                                                                      np.mean(self.losses)))

    def step_begin(self, run_context):
        self.step_time = time.time()

    def step_end(self, run_context):
        cb_params = run_context.original_args()
        step_mseconds = (time.time() - self.step_time) * 1000
        step_loss = cb_params.net_outputs

        if isinstance(step_loss, (tuple, list)) and isinstance(step_loss[0], Tensor):
            step_loss = step_loss[0]
        if isinstance(step_loss, Tensor):
            step_loss = np.mean(step_loss.asnumpy())

        self.losses.append(step_loss)
        cur_step_in_epoch = (cb_params.cur_step_num - 1) % cb_params.batch_num

        print("epoch: [{:3d}/{:3d}], step:[{:5d}/{:5d}], loss:[{:5.3f}/{:5.3f}], time:[{:5.3f}], lr:[{:5.3f}]".format(
            cb_params.cur_epoch_num -
            1, cb_params.epoch_num, cur_step_in_epoch, cb_params.batch_num, step_loss,
            np.mean(self.losses), step_mseconds, self.lr_init[cb_params.cur_step_num - 1]))


class LossTimeMonitorV2(Callback):
    """
    Monitor loss and time version 2.0.
    This version will not show learning rate.

    Args:

    Returns:
        None

    Examples:
        >>> from tinyms.callbacks import LossTimeMonitorV2
        >>>
        >>> LossTimeMonitorV2()
    """

    def __init__(self):
        super(LossTimeMonitorV2, self).__init__()

    def epoch_begin(self, run_context):
        self.losses = []
        self.epoch_time = time.time()

    def epoch_end(self, run_context):
        cb_params = run_context.original_args()

        epoch_mseconds = (time.time() - self.epoch_time) * 1000
        per_step_mseconds = epoch_mseconds / cb_params.batch_num
        print("epoch time: {:5.3f}, per step time: {:5.3f}, avg loss: {:5.3f}".
              format(epoch_mseconds, per_step_mseconds, np.mean(self.losses)), flush=True)

    def step_begin(self, run_context):
        self.step_time = time.time()

    def step_end(self, run_context):
        cb_params = run_context.original_args()
        step_mseconds = (time.time() - self.step_time) * 1000
        step_loss = cb_params.net_outputs
        # arr_lr = cb_params.optimizer.learning_rate.asnumpy()
        # lr = float(np.array2string(arr_lr))

        if isinstance(step_loss, (tuple, list)) and isinstance(step_loss[0], Tensor):
            step_loss = step_loss[0]
        if isinstance(step_loss, Tensor):
            step_loss = np.mean(step_loss.asnumpy())

        self.losses.append(step_loss)
        cur_step_in_epoch = (cb_params.cur_step_num - 1) % cb_params.batch_num

        print("epoch: [{:3d}/{:3d}], step:[{:5d}/{:5d}], loss:[{:5.3f}/{:5.3f}], time:[{:5.3f}]]".format(
            cb_params.cur_epoch_num - 1, cb_params.epoch_num, cur_step_in_epoch, cb_params.batch_num, step_loss,
            np.mean(self.losses), step_mseconds), flush=True)


class BertLossCallBack(Callback):
    """
    Monitor the loss in training.
    If the loss in NAN or INF terminating training.

    Args:
        dataset_size (int): Print loss every times. Default: 1.

    Returns:
        None

    Examples:
        >>> from tinyms.callbacks import BertLossCallBack
        >>>
        >>> BertLossCallBack(dataset_size=1)
    """

    def __init__(self, dataset_size=1):
        super(BertLossCallBack, self).__init__()
        self._dataset_size = dataset_size
    def step_end(self, run_context):
        """
        Print loss after each step
        """
        cb_params = run_context.original_args()
        if self._dataset_size > 0:
            percent, epoch_num = math.modf(cb_params.cur_step_num / self._dataset_size)
            if percent == 0:
                percent = 1
                epoch_num -= 1
            print("epoch: {}, current epoch percent: {}, step: {}, outputs are {}"
                  .format(int(epoch_num), "%.3f" % percent, cb_params.cur_step_num, str(cb_params.net_outputs)))
        else:
            print("epoch: {}, step: {}, outputs are {}".format(cb_params.cur_epoch_num, cb_params.cur_step_num,
                                                               str(cb_params.net_outputs)))







#!/usr/bin/env python3
# encoding: utf-8
# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""setup package."""
import os
import setuptools

package_name = 'tinyms'
version_tag = '0.3.2'
pwd = os.path.dirname(os.path.realpath(__file__))


def _read_file(filename):
    with open(os.path.join(pwd, filename), encoding='UTF-8') as f:
        return f.read()


readme = _read_file('README.md')


def _write_version(file):
    file.write("__version__ = '{}'\n".format(version_tag))


setup_required_package = [
    'wheel>=0.32.0',
    'setuptools>=40.8.0',
]


install_required_package = [
    'numpy>=1.17.0',
    'easydict>=1.9',
    'scipy>=1.5.2,<1.8.0',
    'matplotlib>=3.1.3,<=3.5.1',
    'Pillow>=6.2.0',
    'mindspore>=1.6.0',
    'requests>=2.22.0',
    'flask==2.0.3',
    'python-Levenshtein>=0.10.2',
    'gensim==3.8.1',
    'PyYAML',
    'opencv-python>=4.1.2.30',
    'flask_cors>=3.0.10',
    'wget>=3.2',
    'scikit-learn==1.0.2',
    'tqdm>=4.62.3',
]

test_required_package = [
    'pycocotools>=2.0.0',
]

package_data = {
    'tinyms.hub': [
        'assets/tinyms/*/*.yaml',
    ],
    'tinyms.vision': [
        'configs/transform_config.yaml',
    ]
}
if __name__ == "__main__":
    with open(os.path.join(pwd, package_name, 'version.py'), 'w') as f:
        _write_version(f)

    setuptools.setup(
        name=package_name,
        version=version_tag,
        author='The TinyMS Authors',
        author_email='simple_hlw@163.com',
        url='https://tinyms.readthedocs.io/en/latest/',
        download_url='https://github.com/tinyms-ai/tinyms/tags',
        project_urls={
            'Sources': 'https://github.com/tinyms-ai/tinyms',
            'Issue Tracker': 'https://github.com/tinyms-ai/tinyms/issues',
        },
        description='TinyMS is an Easy-to-Use deep learning development toolkit.',
        long_description="\n\n".join([readme]),
        long_description_content_type="text/markdown",
        packages=setuptools.find_packages(),
        package_data=package_data,
        include_package_data=True,
        python_requires='>=3.7',
        setup_requires=setup_required_package,
        install_requires=install_required_package,
        tests_require=test_required_package,
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.7',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        license='Apache 2.0',
        keywords='machine learning toolkit',
    )



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
.. TinyMS package.

Top-level reference to dtype of common module.
This module also provides Numpy-like interfaces in TinyMS.

Examples:
    >>> import tinyms as ts
    >>>
    >>> print(ts.ones([2, 3]))
    [[1. 1. 1.]
    [1. 1. 1.]]
"""
from .version import __version__
from . import common
from .common import *

__all__ = []
__all__.extend(__version__)
__all__.extend(common.__all__)



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Bert for pretraining."""
import numpy as np

import tinyms as ts
from tinyms import layers
from tinyms import primitives as P
from tinyms import Tensor
from tinyms import Parameter
from tinyms.initializers import TruncatedNormal, initializer

from .bert import Bert

GRADIENT_CLIP_TYPE = 1
GRADIENT_CLIP_VALUE = 1.0

clip_grad = P.MultitypeFuncGraph("clip_grad")

@clip_grad.register("Number", "Number", "Tensor")
def _clip_grad(clip_type, clip_value, grad):
    """
    Clip gradients.

    Inputs:
        clip_type (int): The way to clip, 0 for 'value', 1 for 'norm'.
        clip_value (float): Specifies how much to clip.
        grad (tuple[Tensor]): Gradients.

    Outputs:
        tuple[Tensor], clipped gradients.
    """

    if clip_type not in (0, 1):
        return grad
    dt = P.DType()(grad)

    if clip_type == 0:
        new_grad = P.clip_by_value(grad, P.Cast()(ts.array((-clip_value,)), dt),
                                   P.Cast()(ts.array((clip_value,)), dt))
    else:
        new_grad = layers.ClipByNorm()(grad, P.Cast()(ts.array((clip_value,)), dt))
    return new_grad


class GetMaskedLMOutput(layers.Layer):
    """
    Get masked lm output.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        Tensor, masked lm output.
    """

    def __init__(self, config):
        super(GetMaskedLMOutput, self).__init__()
        self.width = config.hidden_size
        self.reshape = P.Reshape()
        self.gather = P.Gather()

        weight_init = TruncatedNormal(config.initializer_range)
        self.dense = layers.Dense(self.width,
                              config.hidden_size,
                              weight_init=weight_init,
                              activation=config.hidden_act).to_float(config.compute_type)
        self.layernorm = layers.LayerNorm((config.hidden_size,)).to_float(config.compute_type)
        self.output_bias = Parameter(
            initializer(
                'zero',
                config.vocab_size))
        self.matmul = P.MatMul(transpose_b=True)
        self.log_softmax = layers.LogSoftmax(axis=-1)
        self.shape_flat_offsets = (-1, 1)
        self.last_idx = (-1,)
        self.shape_flat_sequence_tensor = (-1, self.width)
        self.seq_length_tensor = Tensor(np.array((config.seq_length,)).astype(np.int32))
        self.cast = P.Cast()
        self.compute_type = config.compute_type
        self.dtype = config.dtype

    def construct(self,
                  input_tensor,
                  output_weights,
                  positions):
        """Get output log_probs"""
        rng = ts.array(ts.arange(P.Shape()(input_tensor)[0]))
        flat_offsets = self.reshape(rng * self.seq_length_tensor, self.shape_flat_offsets)
        flat_position = self.reshape(positions + flat_offsets, self.last_idx)
        flat_sequence_tensor = self.reshape(input_tensor, self.shape_flat_sequence_tensor)
        input_tensor = self.gather(flat_sequence_tensor, flat_position, 0)
        input_tensor = self.cast(input_tensor, self.compute_type)
        output_weights = self.cast(output_weights, self.compute_type)
        input_tensor = self.dense(input_tensor)
        input_tensor = self.layernorm(input_tensor)
        logits = self.matmul(input_tensor, output_weights)
        logits = self.cast(logits, self.dtype)
        logits = logits + self.output_bias
        log_probs = self.log_softmax(logits)
        return log_probs


class GetNextSentenceOutput(layers.Layer):
    """
    Get next sentence output.

    Args:
        config (BertConfig): The config of Bert.

    Returns:
        Tensor, next sentence output.
    """

    def __init__(self, config):
        super(GetNextSentenceOutput, self).__init__()
        self.log_softmax = P.LogSoftmax()
        weight_init = TruncatedNormal(config.initializer_range)
        self.dense = layers.Dense(config.hidden_size, 2,
                              weight_init=weight_init, has_bias=True).to_float(config.compute_type)
        self.dtype = config.dtype
        self.cast = P.Cast()

    def construct(self, input_tensor):
        logits = self.dense(input_tensor)
        logits = self.cast(logits, self.dtype)
        log_prob = self.log_softmax(logits)
        return log_prob


class BertPreTraining(layers.Layer):
    """
    Bert pretraining network.

    Args:
        config (BertConfig): The config of BertModel.
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings.

    Returns:
        Tensor, prediction_scores, seq_relationship_score.
    """

    def __init__(self, config, is_training, use_one_hot_embeddings):
        super(BertPreTraining, self).__init__()
        self.bert = Bert(config, is_training, use_one_hot_embeddings)
        self.cls1 = GetMaskedLMOutput(config)
        self.cls2 = GetNextSentenceOutput(config)

    def construct(self, input_ids, input_mask, token_type_id,
                  masked_lm_positions):
        sequence_output, pooled_output, embedding_table = \
            self.bert(input_ids, token_type_id, input_mask)
        prediction_scores = self.cls1(sequence_output,
                                      embedding_table,
                                      masked_lm_positions)
        seq_relationship_score = self.cls2(pooled_output)
        return prediction_scores, seq_relationship_score


class BertPretrainingLoss(layers.Layer):
    """
    Provide bert pre-training loss.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        Tensor, total loss.
    """

    def __init__(self, config):
        super(BertPretrainingLoss, self).__init__()
        self.vocab_size = config.vocab_size
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, ts.float32)
        self.off_value = Tensor(0.0, ts.float32)
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.reshape = P.Reshape()
        self.last_idx = (-1,)
        self.neg = P.Neg()
        self.cast = P.Cast()

    def construct(self, prediction_scores, seq_relationship_score, masked_lm_ids,
                  masked_lm_weights, next_sentence_labels):
        """Defines the computation performed."""
        label_ids = self.reshape(masked_lm_ids, self.last_idx)
        label_weights = self.cast(self.reshape(masked_lm_weights, self.last_idx), ts.float32)
        one_hot_labels = self.onehot(label_ids, self.vocab_size, self.on_value, self.off_value)

        per_example_loss = self.neg(self.reduce_sum(prediction_scores * one_hot_labels, self.last_idx))
        numerator = self.reduce_sum(label_weights * per_example_loss, ())
        denominator = self.reduce_sum(label_weights, ()) + self.cast(ts.array((1e-5,)), ts.float32)
        masked_lm_loss = numerator / denominator

        # next_sentence_loss
        labels = self.reshape(next_sentence_labels, self.last_idx)
        one_hot_labels = self.onehot(labels, 2, self.on_value, self.off_value)
        per_example_loss = self.neg(self.reduce_sum(
            one_hot_labels * seq_relationship_score, self.last_idx))
        next_sentence_loss = self.reduce_mean(per_example_loss, self.last_idx)

        # total_loss
        total_loss = masked_lm_loss + next_sentence_loss

        return total_loss


class BertNetworkWithLoss(layers.Layer):
    """
    Provide bert pre-training loss through network.

    Args:
        config (BertConfig): The config of BertModel.
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings. Default: False.

    Returns:
        Tensor, the loss of the network.
    """

    def __init__(self, config, is_training, use_one_hot_embeddings=False):
        super(BertNetworkWithLoss, self).__init__()
        self.bert = BertPreTraining(config, is_training, use_one_hot_embeddings)
        self.loss = BertPretrainingLoss(config)
        self.cast = P.Cast()

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights):
        """Get pre-training loss"""
        prediction_scores, seq_relationship_score = \
            self.bert(input_ids, input_mask, token_type_id, masked_lm_positions)
        total_loss = self.loss(prediction_scores, seq_relationship_score,
                               masked_lm_ids, masked_lm_weights, next_sentence_labels)
        return self.cast(total_loss, ts.float32)


class BertTrainOneStepCell(layers.TrainOneStepCell):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        sens (Number): The adjust parameter. Default: 1.0.
    """

    def __init__(self, network, optimizer, sens=1.0):
        super(BertTrainOneStepCell, self).__init__(network, optimizer, sens)
        self.cast = P.Cast()
        self.hyper_map = P.HyperMap()

    def set_sens(self, value):
        self.sens = value

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights):
        """Defines the computation performed."""
        weights = self.weights

        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(ts.array((self.sens,)),
                                                           ts.float32))
        grads = self.hyper_map(P.Partial()(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        grads = self.grad_reducer(grads)
        succ = self.optimizer(grads)
        return P.Depend()(loss, succ)


grad_scale = P.MultitypeFuncGraph("grad_scale")
reciprocal = P.Reciprocal()


@grad_scale.register("Tensor", "Tensor")
def tensor_grad_scale(scale, grad):
    return grad * reciprocal(scale)


_grad_overflow = P.MultitypeFuncGraph("_grad_overflow")
grad_overflow = P.FloatStatus()


@_grad_overflow.register("Tensor")
def _tensor_grad_overflow(grad):
    return grad_overflow(grad)


class BertTrainOneStepWithLossScaleCell(layers.TrainOneStepWithLossScaleCell):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
    """

    def __init__(self, network, optimizer, scale_update_cell=None):
        super(BertTrainOneStepWithLossScaleCell, self).__init__(network, optimizer, scale_update_cell)
        self.cast = P.Cast()
        self.degree = 1
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=ts.float32))

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens
        status, scaling_sens = self.start_overflow_check(loss, scaling_sens)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(scaling_sens,
                                                           ts.float32))
        # apply grad reducer on grads
        grads = self.grad_reducer(grads)
        grads = self.hyper_map(P.Partial()(grad_scale, scaling_sens * self.degree), grads)
        grads = self.hyper_map(P.Partial()(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)

        cond = self.get_overflow_status(status, grads)
        overflow = cond
        if sens is None:
            overflow = self.loss_scaling_manager(self.loss_scale, cond)
        if overflow:
            succ = False
        else:
            succ = self.optimizer(grads)
        ret = (loss, cond, scaling_sens)
        return P.Depend()(ret, succ)


class BertTrainOneStepWithLossScaleCellForAdam(layers.TrainOneStepWithLossScaleCell):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.
    Different from BertTrainOneStepWithLossScaleCell, the optimizer takes the overflow
    condition as input.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
    """
    def __init__(self, network, optimizer, scale_update_cell=None):
        super(BertTrainOneStepWithLossScaleCellForAdam, self).__init__(network, optimizer, scale_update_cell)
        self.cast = P.Cast()
        self.degree = 1
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=ts.float32))

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens

        status, scaling_sens = self.start_overflow_check(loss, scaling_sens)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(scaling_sens,
                                                           ts.float32))
        # apply grad reducer on grads
        grads = self.grad_reducer(grads)
        grads = self.hyper_map(P.Partial()(grad_scale, scaling_sens * self.degree), grads)
        grads = self.hyper_map(P.Partial()(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        cond = self.get_overflow_status(status, grads)
        overflow = cond
        if self.loss_scaling_manager is not None:
            overflow = self.loss_scaling_manager(scaling_sens, cond)
        succ = self.optimizer(grads, overflow)
        ret = (loss, cond, scaling_sens)
        return P.Depend()(ret, succ)

cast = P.Cast()
add_grads = P.MultitypeFuncGraph("add_grads")


@add_grads.register("Tensor", "Tensor")
def _add_grads(accu_grad, grad):
    return accu_grad + cast(grad, ts.float32)

update_accu_grads = P.MultitypeFuncGraph("update_accu_grads")

@update_accu_grads.register("Tensor", "Tensor")
def _update_accu_grads(accu_grad, grad):
    succ = True
    return P.Depend()(succ, P.AssignAdd()(accu_grad, cast(grad, ts.float32)))

accumulate_accu_grads = P.MultitypeFuncGraph("accumulate_accu_grads")

@accumulate_accu_grads.register("Tensor", "Tensor")
def _accumulate_accu_grads(accu_grad, grad):
    succ = True
    return P.Depend()(succ, P.AssignAdd()(accu_grad, cast(grad, ts.float32)))


zeroslike = P.ZerosLike()
reset_accu_grads = P.MultitypeFuncGraph("reset_accu_grads")


@reset_accu_grads.register("Tensor")
def _reset_accu_grads(accu_grad):
    succ = True
    return P.Depend()(succ, P.Assign()(accu_grad, zeroslike(accu_grad)))


class BertTrainAccumulationAllReducePostWithLossScaleLayer(layers.Layer):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    To mimic higher batch size, gradients are accumulated N times before weight update.

    Args:
        network (layers.Layer): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
        accumulation_steps (int): Number of accumulation steps before gradient update. The global batch size =
                                batch_size * accumulation_steps. Default: 1.
    """

    def __init__(self, network, optimizer, scale_update_cell=None, accumulation_steps=1, enable_global_norm=False):
        super(BertTrainAccumulationAllReducePostWithLossScaleLayer, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.accumulation_steps = accumulation_steps
        self.enable_global_norm = enable_global_norm
        self.one = Tensor(np.array([1]).astype(np.int32))
        self.zero = Tensor(np.array([0]).astype(np.int32))
        self.local_step = Parameter(initializer(0, [1], ts.int32))
        self.accu_grads = self.weights.clone(prefix="accu_grads", init='zeros')
        self.accu_overflow = Parameter(initializer(0, [1], ts.int32))
        self.accu_loss = Parameter(initializer(0, [1], ts.float32))

        self.grad = P.GradOperation(get_by_list=True, sens_param=True)
        self.grad_reducer = P.Identity()
        self.degree = 1
        self.overflow_reducer = P.Identity()
        self.cast = P.Cast()
        self.alloc_status = P.NPUAllocFloatStatus()
        self.get_status = P.NPUGetFloatStatus()
        self.clear_status = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.base = Tensor(1, ts.float32)
        self.less_equal = P.LessEqual()
        self.logical_or = P.LogicalOr()
        self.not_equal = P.NotEqual()
        self.select = P.Select()
        self.reshape = P.Reshape()
        self.hyper_map = P.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=ts.float32))

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens
        # alloc status and clear should be right before gradoperation
        init = self.alloc_status()
        init = P.Depend()(init, loss)
        clear_status = self.clear_status(init)
        scaling_sens = P.Depend()(scaling_sens, clear_status)
        # update accumulation parameters
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)
        self.local_step = self.select(is_accu_step, self.local_step + self.one, self.one)
        self.accu_loss = self.select(is_accu_step, self.accu_loss + loss, loss)
        mean_loss = self.accu_loss / self.local_step
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)

        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(scaling_sens,
                                                           ts.float32))

        accu_succ = self.hyper_map(accumulate_accu_grads, self.accu_grads, grads)
        mean_loss = P.Depend()(mean_loss, accu_succ)

        init = P.Depend()(init, mean_loss)
        get_status = self.get_status(init)
        init = P.Depend()(init, get_status)
        flag_sum = self.reduce_sum(init, (0,))
        overflow = self.less_equal(self.base, flag_sum)
        overflow = self.logical_or(self.not_equal(self.accu_overflow, self.zero), overflow)
        accu_overflow = self.select(overflow, self.one, self.zero)
        self.accu_overflow = self.select(is_accu_step, accu_overflow, self.zero)

        if is_accu_step:
            succ = False
        else:
            # apply grad reducer on grads
            grads = self.grad_reducer(self.accu_grads)
            scaling = scaling_sens * self.degree * self.accumulation_steps
            grads = self.hyper_map(P.Partial()(grad_scale, scaling), grads)
            if self.enable_global_norm:
                grads = P.clip_by_global_norm(grads, 1.0, None)
            else:
                grads = self.hyper_map(P.Partial()(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
            accu_overflow = P.Depend()(accu_overflow, grads)
            accu_overflow = self.overflow_reducer(accu_overflow)
            overflow = self.less_equal(self.base, accu_overflow)
            accu_succ = self.hyper_map(reset_accu_grads, self.accu_grads)
            overflow = P.Depend()(overflow, accu_succ)
            overflow = self.reshape(overflow, (()))
            if sens is None:
                overflow = self.loss_scaling_manager(self.loss_scale, overflow)
            if overflow:
                succ = False
            else:
                succ = self.optimizer(grads)

        ret = (mean_loss, overflow, scaling_sens)
        return P.Depend()(ret, succ)


class BertTrainAccumulationAllReduceEachWithLossScaleLayer(layers.Layer):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    To mimic higher batch size, gradients are accumulated N times before weight update.

    Args:
        network (layers.Layer): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
        accumulation_steps (int): Number of accumulation steps before gradient update. The global batch size =
                                  batch_size * accumulation_steps. Default: 1.
    """
    def __init__(self, network, optimizer, scale_update_cell=None, accumulation_steps=1, enable_global_norm=False):
        super(BertTrainAccumulationAllReduceEachWithLossScaleLayer, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.accumulation_steps = accumulation_steps
        self.enable_global_norm = enable_global_norm
        self.one = Tensor(np.array([1]).astype(np.int32))
        self.zero = Tensor(np.array([0]).astype(np.int32))
        self.local_step = Parameter(initializer(0, [1], ts.int32))
        self.accu_grads = self.weights.clone(prefix="accu_grads", init='zeros')
        self.accu_overflow = Parameter(initializer(0, [1], ts.int32))
        self.accu_loss = Parameter(initializer(0, [1], ts.float32))

        self.grad = P.GradOperation(get_by_list=True, sens_param=True)
        self.grad_reducer = P.Identity()
        self.degree = 1
        self.overflow_reducer = P.Identity()
        self.cast = P.Cast()
        self.alloc_status = P.NPUAllocFloatStatus()
        self.get_status = P.NPUGetFloatStatus()
        self.clear_before_grad = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.base = Tensor(1, ts.float32)
        self.less_equal = P.LessEqual()
        self.logical_or = P.LogicalOr()
        self.not_equal = P.NotEqual()
        self.select = P.Select()
        self.reshape = P.Reshape()
        self.hyper_map = P.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=ts.float32))

    @P.add_flags(has_effect=True)
    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens

        # update accumulation parameters
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)
        self.local_step = self.select(is_accu_step, self.local_step + self.one, self.one)
        self.accu_loss = self.select(is_accu_step, self.accu_loss + loss, loss)
        mean_loss = self.accu_loss / self.local_step
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)

        # alloc status and clear should be right before gradoperation
        init = self.alloc_status()
        self.clear_before_grad(init)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(scaling_sens,
                                                           ts.float32))


        accu_grads = self.hyper_map(add_grads, self.accu_grads, grads)
        scaling = scaling_sens * self.degree * self.accumulation_steps
        grads = self.hyper_map(P.Partial()(grad_scale, scaling), accu_grads)
        grads = self.grad_reducer(grads)

        self.get_status(init)
        flag_sum = self.reduce_sum(init, (0,))
        flag_reduce = self.overflow_reducer(flag_sum)
        overflow = self.less_equal(self.base, flag_reduce)
        overflow = self.logical_or(self.not_equal(self.accu_overflow, self.zero), overflow)
        accu_overflow = self.select(overflow, self.one, self.zero)
        self.accu_overflow = self.select(is_accu_step, accu_overflow, self.zero)
        overflow = self.reshape(overflow, (()))

        if is_accu_step:
            succ = False
            accu_succ = self.hyper_map(update_accu_grads, self.accu_grads, accu_grads)
            succ = P.Depend()(succ, accu_succ)
        else:
            if sens is None:
                overflow = self.loss_scaling_manager(self.loss_scale, overflow)
            if overflow:
                succ = False
            else:
                if self.enable_global_norm:
                    grads = P.clip_by_global_norm(grads, 1.0, None)
                else:
                    grads = self.hyper_map(P.Partial()(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)

                succ = self.optimizer(grads)

            accu_succ = self.hyper_map(reset_accu_grads, self.accu_grads)
            succ = P.Depend()(succ, accu_succ)

        ret = (mean_loss, overflow, scaling_sens)
        return P.Depend()(ret, succ)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Learning rate generator"""
import math
import tinyms as ts

__all__ = ['mobilenetv2_lr', 'cyclegan_lr']


def mobilenetv2_lr(global_step, lr_init, lr_end, lr_max, warmup_epochs, total_epochs, steps_per_epoch):
    """
    Generate learning rate for mobilenetv2.

    Args:
       global_step (int): Total steps of the training.
       lr_init (float): Init learning rate.
       lr_end (float): End learning rate.
       lr_max (float): Max learning rate.
       warmup_epochs (int): Number of warmup epochs.
       total_epochs (int): Total epoch of training.
       steps_per_epoch (int): Steps of one epoch.

    Returns:
       Tensor, learning rate.
    """
    lr_each_step = []
    total_steps = steps_per_epoch * total_epochs
    warmup_steps = steps_per_epoch * warmup_epochs
    for i in range(total_steps):
        if i < warmup_steps:
            lr = lr_init + (lr_max - lr_init) * i / warmup_steps
        else:
            lr = lr_end + \
                (lr_max - lr_end) * \
                (1. + math.cos(math.pi * (i - warmup_steps) / (total_steps - warmup_steps))) / 2.
        if lr < 0.0:
            lr = 0.0
        lr_each_step.append(lr)

    current_step = global_step
    lr_each_step = ts.array(lr_each_step, dtype=ts.float32)
    learning_rate = lr_each_step[current_step:]

    return learning_rate


def cyclegan_lr(max_epoch, n_epoch, dataset_size):
    """
    Generate learning rate for cycle_gan.

    Args:
       max_epoch (int): Epoch size for training.
       n_epoch (int): Number of epochs with the initial learning rate.
       dataset_size (int): Total size of dataset.

    Returns:
       Tensor, learning rate.
    """
    n_epochs_decay = max_epoch - n_epoch
    lrs = [0.0002] * dataset_size * n_epoch
    lr_epoch = 0
    for epoch in range(n_epochs_decay):
        lr_epoch = 0.0002 * (n_epochs_decay - epoch) / n_epochs_decay
        lrs += [lr_epoch] * dataset_size
    lrs += [lr_epoch] * dataset_size * (max_epoch - n_epochs_decay - n_epoch)
    return ts.array(lrs, dtype=ts.float32)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""The CityScapes util class for evaluation."""

import numpy as np
from tinyms.data.utils import load_img

# label name and RGB color map.
label2color = {
    'unlabeled': (0, 0, 0),
    'ego vehicle': (0, 0, 0),
    'rectification border': (0, 0, 0),
    'out of roi': (0, 0, 0),
    'static': (0, 0, 0),
    'dynamic': (111, 74, 0),
    'ground': (81, 0, 81),
    'road': (128, 64, 128),
    'sidewalk': (244, 35, 232),
    'parking': (250, 170, 160),
    'rail track': (230, 150, 140),
    'building': (70, 70, 70),
    'wall': (102, 102, 156),
    'fence': (190, 153, 153),
    'guard rail': (180, 165, 180),
    'bridge': (150, 100, 100),
    'tunnel': (150, 120, 90),
    'pole': (153, 153, 153),
    'polegroup': (153, 153, 153),
    'traffic light': (250, 170, 30),
    'traffic sign': (220, 220, 0),
    'vegetation': (107, 142, 35),
    'terrain': (152, 251, 152),
    'sky': (70, 130, 180),
    'person': (220, 20, 60),
    'rider': (255, 0, 0),
    'car': (0, 0, 142),
    'truck': (0, 0, 70),
    'bus': (0, 60, 100),
    'caravan': (0, 0, 90),
    'trailer': (0, 0, 110),
    'train': (0, 80, 100),
    'motorcycle': (0, 0, 230),
    'bicycle': (119, 11, 32),
    'license plate': (0, 0, 142)
}


def fast_hist(a, b, n):
    """
    Fast histogram

    Args:
        a (numpy.ndarray): The flatten train ids of cityscapes_dir image
        b (numpy.ndarray): The flatten train ids of result_dir image
        n (int): The number of cityscapes classes

    Returns:
        Numpy, histogram numpy
    """
    k = np.where((a >= 0) & (a < n))[0]
    bc = np.bincount(n * a[k].astype(int) + b[k], minlength=n**2)
    if len(bc) != n**2:
        # ignore this example if dimension mismatch
        return 0
    return bc.reshape(n, n)


def get_scores(hist):
    """
    Get accuracy scores

    Args:
        hist (numpy.ndarray): Histogram numpy

    Returns:
        Tuple, Mean pixel accuracy, mean per class accuracy, mean per class IoU, per class accuracy and per class IoU
    """
    # Mean pixel accuracy
    acc = np.diag(hist).sum() / (hist.sum() + 1e-12)
    # Per class accuracy
    cl_acc = np.diag(hist) / (hist.sum(1) + 1e-12)
    # Per class IoU
    iu = np.diag(hist) / (hist.sum(1) + hist.sum(0) - np.diag(hist) + 1e-12)
    return acc, np.nanmean(cl_acc), np.nanmean(iu), cl_acc, iu


class CityScapes:
    """
    The CityScapes util class for evaluation.

    Returns:
        CityScapes instance
    """
    def __init__(self):
        self.classes = ['road', 'sidewalk', 'building', 'wall', 'fence',
                        'pole', 'traffic light', 'traffic sign', 'vegetation', 'terrain',
                        'sky', 'person', 'rider', 'car', 'truck',
                        'bus', 'train', 'motorcycle', 'bicycle', 'unlabeled']
        self.color_list = []
        for name in self.classes:
            self.color_list.append(label2color[name])
        self.class_num = len(self.classes)

    def get_id(self, img_path):
        """
        Get train id by img

        Args:
            img_path (str): The path of the image.

        Returns:
            numpy.ndarray, train ids
        """
        img = np.array(load_img(img_path))
        w, h, _ = img.shape
        img_tile = np.tile(img, (1, 1, self.class_num)).reshape(w, h, self.class_num, 3)
        diff = np.abs(img_tile - self.color_list).sum(axis=-1)
        ids = diff.argmin(axis=-1)
        return ids



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cycle GAN network."""

import tinyms as ts
from tinyms import layers
from tinyms.primitives import OnesLike, GradOperation, Fill, DType, Shape, depend
from .resnet import ResNetGenerator
from .unet import UnetGenerator
from .common_net import ConvNormReLU, init_weights


def get_generator(model):
    """
        Get generator by model.

        Args:
            model (str): Currently it should be in [resnet, unet].

        Returns:
            Generator, generator net.

        Raises:
            NotImplementedError: If `model` is not in [resnet, unet].
    """
    if model == "resnet":
        net = ResNetGenerator(in_planes=3, ngf=64, n_layers=9, alpha=0.2,
                              norm_mode='instance', dropout=True, pad_mode='CONSTANT')
        init_weights(net, init_type='normal', init_gain=0.02)
    elif model == "unet":
        net = UnetGenerator(in_planes=3, out_planes=3, ngf=64, n_layers=9,
                            alpha=0.2, norm_mode='instance', dropout=True)
        init_weights(net, init_type='normal', init_gain=0.02)
    else:
        raise NotImplementedError(f'Model {model} not recognized.')
    return net


def get_discriminator():
    """
        Get discriminator net.

        Returns:
            Discriminator, discriminator net.
    """
    net = Discriminator(in_planes=3, ndf=64, n_layers=3, alpha=0.2, norm_mode='instance')
    init_weights(net, init_type='normal', init_gain=0.02)
    return net


class Discriminator(layers.Layer):
    """
    Discriminator of GAN.

    Args:
        in_planes (int): Input channel.
        ndf (int): Output channel.
        n_layers (int): The number of ConvNormReLU blocks.
        alpha (float): LeakyRelu slope. Default: 0.2.
        norm_mode (str): Specifies norm method. The optional values are "batch", "instance".

    Returns:
        Tensor, output tensor.

    Examples:
        >>> Discriminator(3, 64, 3)
    """

    def __init__(self, in_planes=3, ndf=64, n_layers=3, alpha=0.2, norm_mode='batch'):
        super(Discriminator, self).__init__()
        kernel_size = 4
        layer_list = [
            layers.Conv2d(in_planes, ndf, kernel_size, 2, pad_mode='pad', padding=1),
            layers.LeakyReLU(alpha)
        ]
        nf_mult = ndf
        for i in range(1, n_layers):
            nf_mult_prev = nf_mult
            nf_mult = min(2 ** i, 8) * ndf
            layer_list.append(ConvNormReLU(nf_mult_prev, nf_mult, kernel_size, 2, alpha, norm_mode, padding=1))
        nf_mult_prev = nf_mult
        nf_mult = min(2 ** n_layers, 8) * ndf
        layer_list.append(ConvNormReLU(nf_mult_prev, nf_mult, kernel_size, 1, alpha, norm_mode, padding=1))
        layer_list.append(layers.Conv2d(nf_mult, 1, kernel_size, 1, pad_mode='pad', padding=1))
        self.features = layers.SequentialLayer(layer_list)

    def construct(self, x):
        output = self.features(x)
        return output


class Generator(layers.Layer):
    """
    Generator of CycleGAN, return fake_A, fake_B, rec_A, rec_B, identity_A and identity_B.

    Args:
        G_A (layers.Layer): The generator network of domain A to domain B.
        G_B (layers.Layer): The generator network of domain B to domain A.
        use_identity (bool): Use identity loss or not. Default: True.

    Returns:
        Tensors, fake_A, fake_B, rec_A, rec_B, identity_A and identity_B.

    Examples:
        >>> Generator(G_A, G_B)
    """

    def __init__(self, G_A, G_B, use_identity=True):
        super(Generator, self).__init__()
        self.G_A = G_A
        self.G_B = G_B
        self.ones = OnesLike()
        self.use_identity = use_identity

    def construct(self, img_A, img_B):
        """If use_identity, identity loss will be used."""
        fake_A = self.G_B(img_B)
        fake_B = self.G_A(img_A)
        rec_A = self.G_B(fake_B)
        rec_B = self.G_A(fake_A)
        if self.use_identity:
            identity_A = self.G_B(img_A)
            identity_B = self.G_A(img_B)
        else:
            identity_A = self.ones(img_A)
            identity_B = self.ones(img_B)
        return fake_A, fake_B, rec_A, rec_B, identity_A, identity_B


class WithLossCell(layers.Layer):
    """
    Wrap the network with loss function to return generator loss.

    Args:
        network (Layer): The target network to wrap.

    Returns:
       Generator Loss lg
    """

    def __init__(self, network):
        super(WithLossCell, self).__init__(auto_prefix=False)
        self.network = network

    def construct(self, img_A, img_B):
        _, _, lg, _, _, _, _, _, _ = self.network(img_A, img_B)
        return lg


class TrainOneStepG(layers.Layer):
    """
    Encapsulation class of Cycle GAN generator network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        G (layers.Layer): Generator with loss Layer. Note that loss function should have been added.
        generator (layers.Layer): Generator of CycleGAN.
        optimizer (Optimizer): Optimizer for updating the weights.
        sens (Number): The adjust parameter. Default: 1.0.
    """

    def __init__(self, G, generator, optimizer, sens=1.0):
        super(TrainOneStepG, self).__init__(auto_prefix=False)
        self.optimizer = optimizer
        self.G = G
        self.G.set_grad()
        self.G.set_train()
        self.G.D_A.set_grad(False)
        self.G.D_A.set_train(False)
        self.G.D_B.set_grad(False)
        self.G.D_B.set_train(False)
        self.grad = GradOperation(get_by_list=True, sens_param=True)
        self.sens = sens
        self.weights = ts.ParameterTuple(generator.trainable_params())
        self.net = WithLossCell(G)

    def construct(self, img_A, img_B):
        weights = self.weights
        fake_A, fake_B, lg, lga, lgb, lca, lcb, lia, lib = self.G(img_A, img_B)
        sens = Fill()(DType()(lg), Shape()(lg), self.sens)
        grads_g = self.grad(self.net, weights)(img_A, img_B, sens)
        return fake_A, fake_B, depend(lg, self.optimizer(grads_g)), lga, lgb, lca, lcb, lia, lib


class TrainOneStepD(layers.Layer):
    """
    Encapsulation class of Cycle GAN discriminator network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        G (layers.Layer): Generator with loss Layer. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        sens (Number): The adjust parameter. Default: 1.0.
    """

    def __init__(self, D, optimizer, sens=1.0):
        super(TrainOneStepD, self).__init__(auto_prefix=False)
        self.optimizer = optimizer
        self.D = D
        self.D.set_grad()
        self.D.set_train()
        self.grad = GradOperation(get_by_list=True, sens_param=True)
        self.sens = sens
        self.weights = ts.ParameterTuple(D.trainable_params())

    def construct(self, img_A, img_B, fake_A, fake_B):
        weights = self.weights
        ld = self.D(img_A, img_B, fake_A, fake_B)
        sens_d = Fill()(DType()(ld), Shape()(ld), self.sens)
        grads_d = self.grad(self.D, weights)(img_A, img_B, fake_A, fake_B, sens_d)
        return depend(ld, self.optimizer(grads_d))


def get_generator_discriminator(model='resnet'):
    """
    Get G_A, G_B generator network and  D_A, D_B discriminator network.

    Args:
        model: The generator model, currently it should be in [resnet, unet].

    Returns:
        G_A, G_B, D_A, D_B network.

    Examples:
        >>> G_A, G_B, D_A, D_B = cycle_gan('resnet')
    """
    if model not in ['resnet', 'unet']:
        raise NotImplementedError(f'Model {model} not recognized.')

    G_A = get_generator(model)
    G_B = get_generator(model)
    D_A = get_discriminator()
    D_B = get_discriminator()

    return G_A, G_B, D_A, D_B


def cycle_gan(G_A, G_B):
    """
    Get Cycle GAN network.

    Args:
        G_A (layers.Layer): The generator net, currently it should be in [resnet, unet].
        G_B (layers.Layer): The generator net, currently it should be in [resnet, unet].

    Returns:
        Cycle GAN instance.

    Examples:
        >>> gan_net = cycle_gan(G_A, G_B)
    """
    if not isinstance(G_A, layers.Layer) or not isinstance(G_B, layers.Layer):
        raise NotImplementedError(f'G_A and G_B are not the instance of layers.Layer')
    return Generator(G_A, G_B)


def cycle_gan_infer(g_model='resnet'):
    """
    Get Cycle GAN network for predict.

    Args:
        g_model (str): The generator network type, currently it should be in ['resnet', 'unet']. Default: resnet.

    Returns:
        Cycle GAN instance.

    Examples:
        >>> gan_net = cycle_gan(G_A, G_B)
    """
    if g_model not in ['resnet', 'unet']:
        raise NotImplementedError(f'Model {g_model} not recognized.')

    G_A = get_generator(g_model)
    G_B = get_generator(g_model)
    return G_A, G_B



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
'''
Bert finetune and evaluation model script.
'''

from tinyms import layers
from tinyms.initializers import TruncatedNormal
from tinyms import primitives as P
from .bert import Bert


class BertCLSModel(layers.Layer):
    """
    This class is responsible for classification task evaluation, i.e. XNLI(num_labels=3),
    LCQMC(num_labels=2), Chnsenti(num_labels=2). The returned output represents the final
    logits as the results of log_softmax is propotional to that of softmax.
    """

    def __init__(self, config, is_training, num_labels=2, dropout_prob=0.0, use_one_hot_embeddings=False,
                 assessment_method=""):
        super(BertCLSModel, self).__init__()
        if not is_training:
            config.hidden_dropout_prob = 0.0
            config.hidden_probs_dropout_prob = 0.0
        self.bert = Bert(config, is_training, use_one_hot_embeddings)
        self.cast = P.Cast()
        self.weight_init = TruncatedNormal(config.initializer_range)
        self.log_softmax = P.LogSoftmax(axis=-1)
        self.dtype = config.dtype
        self.num_labels = num_labels
        self.dense_1 = layers.Dense(config.hidden_size, self.num_labels, weight_init=self.weight_init,
                                has_bias=True).to_float(config.compute_type)
        self.dropout = layers.Dropout(1 - dropout_prob)
        self.assessment_method = assessment_method

    def construct(self, input_ids, input_mask, token_type_id):
        _, pooled_output, _ = \
            self.bert(input_ids, token_type_id, input_mask)
        cls = self.cast(pooled_output, self.dtype)
        cls = self.dropout(cls)
        logits = self.dense_1(cls)
        logits = self.cast(logits, self.dtype)
        if self.assessment_method != "spearman_correlation":
            logits = self.log_softmax(logits)
        return logits


class BertSquadModel(layers.Layer):
    '''
    This class is responsible for SQuAD
    '''

    def __init__(self, config, is_training, num_labels=2, dropout_prob=0.0, use_one_hot_embeddings=False):
        super(BertSquadModel, self).__init__()
        if not is_training:
            config.hidden_dropout_prob = 0.0
            config.hidden_probs_dropout_prob = 0.0
        self.bert = Bert(config, is_training, use_one_hot_embeddings)
        self.weight_init = TruncatedNormal(config.initializer_range)
        self.dense1 = layers.Dense(config.hidden_size, num_labels, weight_init=self.weight_init,
                               has_bias=True).to_float(config.compute_type)
        self.num_labels = num_labels
        self.dtype = config.dtype
        self.log_softmax = P.LogSoftmax(axis=1)
        self.is_training = is_training

    def construct(self, input_ids, input_mask, token_type_id):
        sequence_output, _, _ = self.bert(input_ids, token_type_id, input_mask)
        batch_size, seq_length, hidden_size = P.Shape()(sequence_output)
        sequence = P.Reshape()(sequence_output, (-1, hidden_size))
        logits = self.dense1(sequence)
        logits = P.Cast()(logits, self.dtype)
        logits = P.Reshape()(logits, (batch_size, seq_length, self.num_labels))
        logits = self.log_softmax(logits)
        return logits


class BertNERModel(layers.Layer):
    """
    This class is responsible for sequence labeling task evaluation, i.e. NER(num_labels=11).
    The returned output represents the final logits as the results of log_softmax is propotional to that of softmax.
    """

    def __init__(self, config, is_training, num_labels=11, use_crf=False, dropout_prob=0.0,
                 use_one_hot_embeddings=False):
        super(BertNERModel, self).__init__()
        if not is_training:
            config.hidden_dropout_prob = 0.0
            config.hidden_probs_dropout_prob = 0.0
        self.bert = Bert(config, is_training, use_one_hot_embeddings)
        self.cast = P.Cast()
        self.weight_init = TruncatedNormal(config.initializer_range)
        self.log_softmax = P.LogSoftmax(axis=-1)
        self.dtype = config.dtype
        self.num_labels = num_labels
        self.dense_1 = layers.Dense(config.hidden_size, self.num_labels, weight_init=self.weight_init,
                                has_bias=True).to_float(config.compute_type)
        self.dropout = layers.Dropout(1 - dropout_prob)
        self.reshape = P.Reshape()
        self.shape = (-1, config.hidden_size)
        self.use_crf = use_crf
        self.origin_shape = (-1, config.seq_length, self.num_labels)

    def construct(self, input_ids, input_mask, token_type_id):
        """Return the final logits as the results of log_softmax."""

        sequence_output, _, _ = \
            self.bert(input_ids, token_type_id, input_mask)
        seq = self.dropout(sequence_output)
        seq = self.reshape(seq, self.shape)
        logits = self.dense_1(seq)
        logits = self.cast(logits, self.dtype)
        if self.use_crf:
            return_value = self.reshape(logits, self.origin_shape)
        else:
            return_value = self.log_softmax(logits)
        return return_value


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import math
import numpy as np

import tinyms as ts
from tinyms import layers, Tensor
from tinyms import primitives as P
from tinyms import ParameterTuple, context, Parameter
from tinyms.initializers import initializer


STACK_LSTM_DEVICE = ["CPU"]


# Initialize short-term memory (h) and long-term memory (c) to 0
def lstm_default_state(batch_size, hidden_size, num_layers, bidirectional):
    """init default input."""
    num_directions = 2 if bidirectional else 1
    h = ts.zeros((num_layers * num_directions, batch_size, hidden_size))
    c = ts.zeros((num_layers * num_directions, batch_size, hidden_size))
    return h, c


def stack_lstm_default_state(batch_size, hidden_size, num_layers, bidirectional):
    """init default input."""
    num_directions = 2 if bidirectional else 1

    h_list = c_list = []
    for _ in range(num_layers):
        h_list.append(ts.zeros((num_directions, batch_size, hidden_size)))
        c_list.append(ts.zeros((num_directions, batch_size, hidden_size)))
    h, c = tuple(h_list), tuple(c_list)
    return h, c


class StackLSTM(layers.Layer):
    """
    Stack multi-layers LSTM together.
    """

    def __init__(self,
                 input_size,
                 hidden_size,
                 num_layers=1,
                 has_bias=True,
                 batch_first=False,
                 dropout=0.0,
                 bidirectional=False):
        super(StackLSTM, self).__init__()
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.transpose = P.Transpose()

        # direction number
        num_directions = 2 if bidirectional else 1

        # input_size list
        input_size_list = [input_size]
        for i in range(num_layers - 1):
            input_size_list.append(hidden_size * num_directions)

        # layers
        layer_list = []
        for i in range(num_layers):
            layer_list.append(layers.LSTMCell(input_size=input_size_list[i],
                                              hidden_size=hidden_size,
                                              has_bias=has_bias,
                                              batch_first=batch_first,
                                              bidirectional=bidirectional,
                                              dropout=dropout))

        # weights
        weights = []
        for i in range(num_layers):
            # weight size
            weight_size = (input_size_list[i] + hidden_size) * num_directions * hidden_size * 4
            if has_bias:
                bias_size = num_directions * hidden_size * 4
                weight_size = weight_size + bias_size

            # numpy weight
            stdv = 1 / math.sqrt(hidden_size)
            w_np = np.random.uniform(-stdv, stdv, (weight_size, 1, 1)).astype(np.float32)

            # lstm weight
            weights.append(Parameter(initializer(Tensor(w_np), w_np.shape), name="weight" + str(i)))

        #
        self.lstm = layer_list
        self.weight = ParameterTuple(tuple(weights))

    def construct(self, x, hx):
        """construct"""
        h, c = hx
        if self.batch_first:
            x = self.transpose(x, (1, 0, 2))
        # stack lstm
        hn = cn = None
        for i in range(self.num_layers):
            x, hn, cn, _, _ = self.lstm[i](x, h[i], c[i], self.weight[i])
        if self.batch_first:
            x = self.transpose(x, (1, 0, 2))
        return x, (hn, cn)


class SentimentNet(layers.Layer):
    """Sentiment network structure."""

    def __init__(self,
                 vocab_size,
                 embed_size,
                 num_hiddens,
                 num_layers,
                 bidirectional,
                 num_classes,
                 weight,
                 batch_size):
        super(SentimentNet, self).__init__()
        # Map words to vectors
        self.embedding = layers.Embedding(vocab_size,
                                          embed_size,
                                          embedding_table=weight)
        self.embedding.embedding_table.requires_grad = False
        self.trans = P.Transpose()
        self.perm = (1, 0, 2)

        if context.get_context("device_target") in STACK_LSTM_DEVICE:
            # stack lstm by user
            self.encoder = StackLSTM(input_size=embed_size,
                                     hidden_size=num_hiddens,
                                     num_layers=num_layers,
                                     has_bias=True,
                                     bidirectional=bidirectional,
                                     dropout=0.0)
            self.h, self.c = stack_lstm_default_state(batch_size, num_hiddens, num_layers, bidirectional)
        else:
            # standard lstm
            self.encoder = layers.LSTM(input_size=embed_size,
                                       hidden_size=num_hiddens,
                                       num_layers=num_layers,
                                       has_bias=True,
                                       bidirectional=bidirectional,
                                       dropout=0.0)
            self.h, self.c = lstm_default_state(batch_size, num_hiddens, num_layers, bidirectional)

        self.concat = P.Concat(1)
        if bidirectional:
            self.decoder = layers.Dense(num_hiddens * 4, num_classes)
        else:
            self.decoder = layers.Dense(num_hiddens * 2, num_classes)

    def construct(self, inputs):
        # input：(64,500,300)
        embeddings = self.embedding(inputs)
        embeddings = self.trans(embeddings, self.perm)
        output, _ = self.encoder(embeddings, (self.h, self.c))
        # states[i] size(64,200)  -> encoding.size(64,400)
        encoding = self.concat((output[0], output[499]))
        outputs = self.decoder(encoding)
        return outputs


def sentimentnet(
    vocab_size,
    embed_size,
    num_hiddens,
    num_layers,
    bidirectional,
    num_classes,
    weight,
    batch_size
):
    """Sentiment network structure."""

    return SentimentNet(
        vocab_size,
        embed_size,
        num_hiddens,
        num_layers,
        bidirectional,
        num_classes,
        weight,
        batch_size
    )



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
from mindspore.dataset.text.transforms import Lookup, JiebaTokenizer, UnicodeCharTokenizer, Ngram, \
    WordpieceTokenizer, TruncateSequencePair, \
    ToNumber, SlidingWindow, SentencePieceTokenizer, PythonTokenizer
from mindspore.dataset.text.utils import to_str, to_bytes, Vocab, SentencePieceVocab, SentencePieceModel, \
    SPieceTokenizerOutType, SPieceTokenizerLoadType
from mindspore.dataset.transforms.c_transforms import *


text_transform = [
    "Lookup",
    "JiebaTokenizer",
    "UnicodeCharTokenizer",
    "Ngram",
    "WordpieceTokenizer",
    "TruncateSequencePair",
    "ToNumber",
    "SlidingWindow",
    "SentencePieceTokenizer",
    "PythonTokenizer"
]

text_utils = [
    "to_str",
    "to_bytes",
    "Vocab",
    "SentencePieceVocab",
    "SentencePieceModel",
    "SPieceTokenizerOutType",
    "SPieceTokenizerLoadType"
]

common_trans = [
    'Compose',
    'Concatenate',
    'Duplicate',
    'Fill',
    'Mask',
    'OneHot',
    'PadEnd',
    'RandomApply',
    'RandomChoice',
    'Slice',
    'TypeCast',
    'Unique',
]

__all__ = []
__all__.extend(text_transform)
__all__.extend(text_utils)
__all__.extend(common_trans)


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Servable functions at the server part"""
import base64
import os
import json
import sys
from io import BytesIO

import numpy as np
from PIL import Image
import cv2

import tinyms as ts
from tinyms import model
from tinyms.utils.predict.predict import cyclegan_predict
from tinyms.vision import mnist_transform, cifar10_transform, imagefolder_transform, \
    voc_transform, shanshui_tranform, cyclegan_transform


serving_path = '/etc/tinyms/serving/'

if os.path.exists('temp.json'):
    with open('temp.json', 'r') as f:
        data = json.load(f)
        serving_path = data.get('serving_path', serving_path)

servable_path = os.path.join(serving_path, 'servable.json')

model_checker = {
    "lenet5": model.lenet5,
    "resnet50": model.resnet50,
    "mobilenetv2": model.mobilenetv2,
    "ssd300": model.ssd300_mobilenetv2,
    "cycle_gan": model.cycle_gan_infer
}

transform_checker = {
    'mnist': mnist_transform,
    'cifar10': cifar10_transform,
    'imagenet2012': imagefolder_transform,
    'voc': voc_transform,
    'shanshui': shanshui_tranform,
    'cityscape': cyclegan_transform
}


def draw_boxes_in_image(bbox_data, img):
    for i in range(len(bbox_data)):
        x = int(bbox_data[i]['bbox'][0])
        y = int(bbox_data[i]['bbox'][1])
        w = int(bbox_data[i]['bbox'][2])
        h = int(bbox_data[i]['bbox'][3])
        print("(x, y, w, h): {}, {}, {}, {}".format(x, y, w, h))
        cv2.rectangle(img, (x, y), (x+w, y+h), (0, 255, 0), 3)
        score = round(bbox_data[i]['score'], 3)
        species = bbox_data[i]['category_id']
        text = species + ', ' + str(score)
        cv2.putText(img, text, (x, y), cv2.FONT_HERSHEY_PLAIN, 3, (0, 0, 255), 3)
    return img


def numpy2base64(img_np):
    img_np = Image.fromarray(img_np)
    output_buffer = BytesIO()
    img_np.save(output_buffer, format="JPEG")
    byte_data = output_buffer.getvalue()
    base64_str = "data:image/jpeg;base64," + str(base64.b64encode(byte_data), encoding="utf-8")
    return base64_str


def servable_search(name=None):
    """
    Check whether the servable json exists and whether the content is valid.

    If the servable exits and the content is valid, the servable values will be returned, otherwise returns the servable list

    Args:
        name (str): servable name

    Returns:
        A string of servable values will be returned if servable json exists, otherwise error message.

    Examples:
        >>> # In the server part, before running the predict function, servable_search is called to check and get the result.
        >>> res = servable_search(servable_name)
        >>> servable = res['servables'][0]
        >>> res = predict(instance, servable_name, servable['model'], strategy)
    """

    # Check if servable_path existed
    if not os.path.exists(servable_path):
        err_msg = "Servable NOT found in " + servable_path
        return {"status": 1, "err_msg": err_msg}

    with open(servable_path, 'r') as f:
        servable_list = json.load(f)
    if name is not None:
        # check if servable name is valid
        def servable_exist(name):
            for servable in servable_list:
                if name in servable.values():
                    return servable
            return None

        servable = servable_exist(name)
        if servable is None:
            err_msg = "Servable name NOT supported!"
            return {"status": 1, "err_msg": err_msg}
        else:
            return {"status": 0, "servables": [servable]}
    else:
        return {"status": 0, "servables": servable_list}


def predict(instance, servable_name, servable_model, strategy):
    """
    Predict the result based on the input data.

    A network will be constructed based on the input and servable data, then load the checkpoint and do the predict.

    Args:
        instance (dict): the dict of input image after transformation, with keys of `shape`, `dtype` and `data`(Image object).
        servable_name (str): servable name
        servable_model (str): name of the model
        strategy (str): output strategy, usually select between `TOP1_CLASS` and `TOP5_CLASS`, for cyclegan, select between `gray2color` and `color2gray`

    Returns:
        The dict object of predicted result after post process.

    Examples:
        >>> # In the server part, after servable_search
        >>> res = predict(instance, servable_name, servable['model'], strategy)
        >>> return jsonify(res)
    """

    # check if servable model name is valid
    model_name = servable_model['name']
    net_func = model_checker.get(model_name)
    if net_func is None:
        err_msg = "Currently model_name only supports " + str(list(model_checker.keys())) + "!"
        return {"status": 1, "err_msg": err_msg}

    # check if model_format is valid
    model_format = servable_model['format']
    if model_format not in ("ckpt"):
        err_msg = "Currently model_format only supports `ckpt`!"
        return {"status": 1, "err_msg": err_msg}

    # parse the input data
    input_data = ts.array(json.loads(instance['data']), dtype=instance['dtype'])

    if model_name == "cycle_gan":
        g_model = servable_model['g_model']
        if strategy == 'gray2color':
            # build the network
            G_generator, _ = net_func(g_model=g_model)
            ckpt_name = 'G_A'

        elif strategy == 'color2gray':
            _, G_generator = net_func(g_model=g_model)
            ckpt_name = 'G_B'
        else:
            err_msg = "Currently cycle_gan strategy only supports `gray2color` and `color2gray`!"
            return {"status": 1, "err_msg": err_msg}
        ckpt_path = os.path.join(serving_path, servable_name, ckpt_name + "." + model_format)
        data = cyclegan_predict(G_generator, input_data, ckpt_path)
    else:
        # build the network
        class_num = servable_model['class_num']
        net = net_func(class_num=class_num, is_training=False)
        serve_model = model.Model(net)

        # load checkpoint
        ckpt_path = os.path.join(serving_path, servable_name, model_name + "." + model_format)
        if not os.path.isfile(ckpt_path):
            err_msg = "The model path " + ckpt_path + " not exist!"
            return {"status": 1, "err_msg": err_msg}
        serve_model.load_checkpoint(ckpt_path)

        # execute the network to perform model prediction
        output = serve_model.predict(ts.expand_dims(input_data, 0))

        data = (ts.concatenate((output[0], output[1]), axis=-1).asnumpy() if model_name == "ssd300"
                else output.asnumpy())
    return {
        "status": 0,
        "instance": {
            "shape": data.shape,
            "dtype": data.dtype.name,
            "data": json.dumps(data.tolist())
        }
    }


def web_predict(instance, servable_name, servable_model, dataset_name, strategy):
    """
    Predict the result based on the input data.

    A network will be constructed based on the input and servable data, then load the checkpoint and do the predict.

    Args:
        instance (dict): the dict of input image after transformation, with keys of `shape`, `dtype` and `data`(Image object).
        servable_name (str): servable name
        servable_model (str): name of the model
        strategy (str): output strategy, usually select between `TOP1_CLASS` and `TOP5_CLASS`, for cyclegan, select between `gray2color` and `color2gray`

    Returns:
        The dict object of predicted result after post process.

    Examples:
        >>> # In the server part, after servable_search
        >>> res = web_predict(instance, servable_name, servable['model'], strategy)
        >>> return jsonify(res)
    """

    # check if servable model name is valid
    model_name = servable_model['name']
    net_func = model_checker.get(model_name)
    if net_func is None:
        err_msg = "Currently model_name only supports " + str(list(model_checker.keys())) + "!"
        return {"status": 1, "err_msg": err_msg}

    # check if model_format is valid
    model_format = servable_model['format']
    if model_format not in ("ckpt"):
        err_msg = "Currently model_format only supports `ckpt`!"
        return {"status": 1, "err_msg": err_msg}

    # Check if dataset supports
    trans_func = transform_checker.get(dataset_name)
    if trans_func is None:
        print("Currently dataset_name only supports {}!".format(list(transform_checker.keys())))
        sys.exit(0)

    # process the original data
    ori_img = np.array(json.loads(instance['data']), dtype=instance['dtype'])
    if dataset_name in ['mnist']:
        image = trans_func(ori_img)
    else:
        cvt_image = cv2.cvtColor(ori_img, cv2.COLOR_BGR2RGB)
        image = trans_func(cvt_image)

    input_data = ts.array(image.tolist(), dtype=image.dtype.name)

    res_msg = ''
    if model_name == "cycle_gan":
        g_model = servable_model['g_model']
        if strategy == 'gray2color':
            # build the network
            G_generator, _ = net_func(g_model=g_model)
            ckpt_name = 'G_A'

        elif strategy == 'color2gray':
            _, G_generator = net_func(g_model=g_model)
            ckpt_name = 'G_B'
        else:
            err_msg = "Currently cycle_gan strategy only supports `gray2color` and `color2gray`!"
            return {"status": 1, "err_msg": err_msg}
        ckpt_path = os.path.join(serving_path, servable_name, ckpt_name + "." + model_format)
        out_img = cyclegan_predict(G_generator, input_data, ckpt_path)
        res_msg = '原图使用{}风格迁移效果'.format(strategy)
        data = numpy2base64(out_img)
    else:
        # build the network
        class_num = servable_model['class_num']
        net = net_func(class_num=class_num, is_training=False)
        serve_model = model.Model(net)

        # load checkpoint
        ckpt_path = os.path.join(serving_path, servable_name, model_name + "." + model_format)
        if not os.path.isfile(ckpt_path):
            err_msg = "The model path " + ckpt_path + " not exist!"
            return {"status": 1, "err_msg": err_msg}
        serve_model.load_checkpoint(ckpt_path)

        # execute the network to perform model prediction
        output = serve_model.predict(ts.expand_dims(input_data, 0))

        if model_name == "ssd300":
            output_np = (ts.concatenate((output[0], output[1]), axis=-1).asnumpy())
            ih, iw, _ = instance['shape']
            bbox_data = trans_func.postprocess(output_np, (ih, iw), strategy)
            print(bbox_data)
            bbox_num = len(bbox_data)
            if not bbox_num:
                err_msg = "抱歉！未检测到任何种类，无法标注。"
                return {"status": 1, "err_msg": err_msg}
            out_img = draw_boxes_in_image(bbox_data, ori_img)
            max_det = max(bbox_data, key=lambda k: k['score'])
            max_score = max_det['score']
            category = bbox_data[bbox_data.index(max_det)]['category_id']
            res_msg = '图中共标注了：{}个框，其中物种{}的得分最高, 为{}。'.format(bbox_num, category, round(max_score, 3))
            data = numpy2base64(cv2.cvtColor(out_img, cv2.COLOR_BGR2RGB))
        else:
            output_np = output.asnumpy()
            res_msg = trans_func.postprocess(output_np, strategy)
            data = numpy2base64(ori_img)

    res = {
        "status": 0,
        "instance": {
            "res_msg": res_msg,
            "data": data
        }
    }
    return res
